# -*- coding: utf-8 -*-
# Copyright (c) 2012 - 2014 the GPy Austhors (see AUTHORS.txt)
# Licensed under the BSD 3-clause license (see LICENSE.txt)
"""Custom version of GPCoregionalizedRegression with set_XY that works.
Needs to be used until the upstream PR gets merged"""
import numpy as np
from GPy import kern, util
from GPy.core import GP
from paramz import ObsAr

__all__ = ["GPCoregionalizedRegression"]


class GPCoregionalizedRegression(GP):  # pylint:disable=too-many-ancestors
    """
    Gaussian Process model for heteroscedastic multioutput regression
    This is a thin wrapper around the models.GP class, with a set of sensible defaults
    :param X_list: list of input observations corresponding to each output
    :type X_list: list of numpy arrays
    :param Y_list: list of observed values related to the different noise models
    :type Y_list: list of numpy arrays
    :param kernel: a GPy kernel ** Coregionalized, defaults to RBF ** Coregionalized
    :type kernel: None | GPy.kernel defaults
    :likelihoods_list: a list of likelihoods, defaults to list of Gaussian likelihoods
    :type likelihoods_list: None | a list GPy.likelihoods
    :param name: model name
    :type name: string
    :param W_rank: number tuples of the corregionalization parameters 'W'
        (see coregionalize kernel documentation)
    :type W_rank: integer
    :param kernel_name: name of the kernel
    :type kernel_name: string
    """

    def __init__(  # pylint:disable=too-many-arguments
        self,
        X_list,
        Y_list,
        kernel=None,
        normalizer=None,
        likelihoods_list=None,
        name="GPCR",
        W_rank=1,
        kernel_name="coreg",
    ):

        # Input and Output
        (
            X,  # pylint:disable=invalid-name
            Y,  # pylint:disable=invalid-name
            self.output_index,
        ) = util.multioutput.build_XY(X_list, Y_list)
        Ny = len(Y_list)  # pylint:disable=invalid-name

        # Kernel
        if kernel is None:
            kernel = kern.RBF(X.shape[1] - 1)

            kernel = util.multioutput.ICM(
                input_dim=X.shape[1] - 1,
                num_outputs=Ny,
                kernel=kernel,
                W_rank=W_rank,
                name=kernel_name,
            )

        # Likelihood
        likelihood = util.multioutput.build_likelihood(
            Y_list, self.output_index, likelihoods_list
        )

        super(  # pylint:disable=super-with-arguments
            GPCoregionalizedRegression, self
        ).__init__(
            X,
            Y,
            kernel,
            likelihood,
            Y_metadata={"output_index": self.output_index},
            normalizer=normalizer,
        )

    def set_XY(self, X=None, Y=None):
        if isinstance(X, list):
            X, _, self.output_index = util.multioutput.build_XY(X, None)
        if isinstance(Y, list):
            _, Y, self.output_index = util.multioutput.build_XY(Y, Y)

        self.update_model(False)
        if Y is not None:
            if self.normalizer is not None:
                self.normalizer.scale_by(Y)
                self.Y_normalized = ObsAr(self.normalizer.normalize(Y))
                self.Y = Y
            else:
                self.Y = ObsAr(Y)
                self.Y_normalized = self.Y
        if X is not None:
            self.X = ObsAr(X)

        self.Y_metadata = {
            "output_index": self.output_index,
            "trials": np.ones(self.output_index.shape),
        }

        self.update_model(True)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Implements a PAL class for GBDT models which can predict uncertainity intervals
when used with quantile loss.
For an example of GBDT with quantile loss see
Jablonka, Kevin Maik; Moosavi, Seyed Mohamad; Asgari, Mehrdad; Ireland, Christopher;
Patiny, Luc; Smit, Berend (2020): A Data-Driven Perspective on
the Colours of Metal-Organic Frameworks. ChemRxiv. Preprint.
https://doi.org/10.26434/chemrxiv.13033217.v1

For general information about quantile regression
see https://en.wikipedia.org/wiki/Quantile_regression

Note that the scaling of the hyperrectangles has been derived
for GPR models (with RBF kernels).
"""
import math
from functools import partial

import numpy as np

from .pal_base import PALBase
from .validate_inputs import validate_gbdt_models, validate_interquartile_scaler

__all__ = ["PALGBDT"]


def _train_model_picklable(i, models, design_space, objectives, sampled):
    model = models[i]
    objective = math.floor(i / 3)
    model.fit(
        design_space[sampled[:, objective], :],
        objectives[sampled[:, objective], objective].ravel(),
    )
    return model


class PALGBDT(PALBase):
    """PAL class for a list of LightGBM GBDT models"""

    def __init__(self, *args, **kwargs):
        """Construct the PALGBDT instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (List[Iterable[LGBMRegressor, LGBMRegressor, LGBMRegressor]]:
                Machine learning models. You need to provide a list of iterables.
                One iterable per objective and every iterable contains three
                LGBMRegressors. The first one for the lower uncertainty limits,
                the middle one for the median and the last one for the upper limit.
                To create appropriate models, you need to use the quantile loss.
                If you want to parallelize training, we recommend that you use
                the LightGBM parallelization and fit the models for the different
                objectives in serial fashion.s
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            interquartile_scaler (float, optional): Used to convert the difference
                between the upper and lower quantile into a standard deviation.
                This, is std = (up-low)/interquartile_scaler. Defaults to 1.35,
                following Wan, X., Wang, W., Liu, J. et al.
                Estimating the sample mean and standard deviation from the sample size,
                median, range and/or interquartile range.
                BMC Med Res Methodol 14, 135 (2014).
                https://doi.org/10.1186/1471-2288-14-135

        """
        self.interquartile_scaler = validate_interquartile_scaler(
            kwargs.pop("interquartile_scaler", 1.35)
        )

        super().__init__(*args, **kwargs)

        self.models = validate_gbdt_models(self.models, self.ndim)

    def _set_data(self):
        pass

    def _train(self):
        models_flat = []
        # not using list(sum(self.models, ())) such that we do not
        # have to assume the type
        for model_tuples in self.models:
            for model in model_tuples:
                model.n_jobs = 1
                model.nthreads = 1
                models_flat.append(model)

        train_single_partial = partial(
            _train_model_picklable,
            models=models_flat,
            design_space=self.design_space,
            objectives=self.y,
            sampled=self.sampled,
        )
        models = []

        for model in map(train_single_partial, range(len(models_flat))):
            models.append(model)

        model_tuples = []
        for i in range(0, len(models), 3):
            chunk = models[i : i + 3]
            model_tuples.append(chunk)
        self.models = model_tuples

    def _predict(self):
        means, stds = [], []
        for model_tuple in self.models:
            mean = model_tuple[1].predict(self.design_space)
            upper = model_tuple[2].predict(self.design_space)
            lower = model_tuple[0].predict(self.design_space)
            std = np.abs(upper - lower) / self.interquartile_scaler

            means.append(mean.reshape(-1, 1))
            stds.append(std.reshape(-1, 1))

        self.means = np.hstack(means)
        self.std = np.hstack(stds)

    def _set_hyperparameters(self):
        # ToDo
        pass



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Methods to validate inputs for the PAL classes"""
import collections
import warnings
from copy import deepcopy
from typing import Any, Iterable, List, Sequence, Union

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV

__all__ = [
    "base_validate_models",
    "validate_beta_scale",
    "validate_coef_var",
    "validate_coregionalized_gpy",
    "validate_delta",
    "validate_epsilon",
    "validate_gbdt_models",
    "validate_goals",
    "validate_gpy_model",
    "validate_interquartile_scaler",
    "validate_ndim",
    "validate_njobs",
    "validate_nt_models",
    "validate_number_models",
    "validate_optimizers",
    "validate_positive_integer_list",
    "validate_sklearn_gpr_models",
]


def validate_ndim(ndim: Any) -> int:
    """Make sure that the number of dimensions makes sense

    Args:
        ndim (Any): number of dimensions

    Raises:
        ValueError: If the number of dimensions is not an integer
        ValueError: If the number of dimensions is not greater than 0

    Returns:
        int: the number of dimensions
    """
    if not isinstance(ndim, int):
        raise ValueError("The number of dimensions, ndim, must be a positive integer")

    if ndim <= 0:
        raise ValueError("ndmin must be greater than 0")

    return ndim


def validate_delta(delta: Any) -> float:
    """Make sure that delta is in a reasonable range

    Args:
        delta (Any): Delta hyperparameter

    Raises:
        ValueError: Delta must be in [0,1].

    Returns:
        float: delta
    """
    if (delta > 1) | (delta < 0):
        raise ValueError("The delta values must be in [0,1]")

    return delta


def validate_beta_scale(beta_scale: Any) -> float:
    """

    Args:
        beta_scale (Any): scaling factor for beta

    Raises:
        ValueError: If beta is smaller than 0

    Returns:
        float: scaling factor for beta
    """
    if beta_scale < 0:
        raise ValueError("The beta_scale values must be positive")

    return beta_scale


def validate_epsilon(epsilon: Any, ndim: int) -> np.ndarray:
    """Validate epsilon and return a np.array

    Args:
        epsilon (Any): Epsilon hyperparameter
        ndim (int): Number of dimensions/objectives

    Raises:
        ValueError: If epsilon is a list there must be one float per dimension
        ValueError: Epsilon must be in [0,1]
        ValueError: If epsilon is an array there must be one float per dimension

    Returns:
        np.ndarray:  Array of one epsilon per objective
    """
    if isinstance(epsilon, list):
        if len(epsilon) != ndim:
            raise ValueError(
                "If epsilon is provided as a list,\
                     there must be one float per dimension"
            )

        for value in epsilon:
            if (value > 1) | (value < 0):
                raise ValueError("The epsilon values must be in [0,1]")
        return np.array(epsilon)

    if isinstance(epsilon, np.ndarray):
        if len(epsilon) != ndim:
            raise ValueError(
                "If epsilon is provided as a array,\
                     there must be one float per dimension"
            )

        for value in epsilon:
            if (value > 1) | (value < 0):
                raise ValueError("The epsilon values must be in [0,1]")
        return epsilon

    if (epsilon > 1) | (epsilon < 0):
        raise ValueError("The epsilon values must be in [0,1]")

    warnings.warn(
        """Only one epsilon value provided,
will automatically expand to use the same value in every dimension""",
        UserWarning,
    )
    return np.array([epsilon] * ndim)


def validate_goals(  # pylint:disable=too-many-branches
    goals: Any, ndim: int
) -> np.ndarray:
    """Create a valid array of goals. 1 for maximization, -1
        for objectives that are to be minimized.

    Args:
        goals (Any): List of goals,
            typically provideded as strings 'max' for maximization
            and 'min' for minimization
        ndim (int): number of dimensions

    Raises:
        ValueError: If goals is a list and the length is not equal to ndim
        ValueError: If goals is a list and the elements
            are not strings 'min', 'max' or -1 and 1

    Returns:
        np.ndarray: Array of -1 and 1
    """
    if goals is None:
        warnings.warn(
            "No goals provided, will assume that every dimension should be maximized",
            UserWarning,
        )

        return np.array([1] * ndim)
    if isinstance(goals, list):
        if len(goals) != ndim:
            raise ValueError("If goals is a list, the length must be equal to the ndim")
        for goal in goals:
            if not isinstance(goal, str) | (goal != 1) | (goal != -1):
                raise ValueError("If goals is a list, it must be a list of strings")

        clean_goals = []
        for goal in goals:
            if isinstance(goal, str):
                if "max" in goal.lower():
                    clean_goals.append(1)
                elif "min" in goal.lower():
                    clean_goals.append(-1)
                else:
                    raise ValueError("The strings in the goals list must be min or max")
            elif isinstance(goal, int):
                if goal == 1:
                    clean_goals.append(1)
                elif goal == -1:
                    clean_goals.append(-1)
                else:
                    raise ValueError("The ints in the goals list must be 1 or -1")

        assert len(clean_goals) == ndim
        return np.array(clean_goals)

    raise ValueError(
        "Goal can be set to None or must be a list of strings\
             or -1 and 1 of length equal to ndim"
    )


def base_validate_models(models: Any) -> list:
    """Currently no validation as the predict and train function
    are implemented independet of the base class"""
    if models:
        return models

    raise ValueError("You must provide some models to initialize pyepal")


def validate_number_models(models: Any, ndim: int):
    """Make sure that there are as many models as objectives

    Args:
        models (Any): List of models
        ndim (int): Number of objectives

    Raises:
        ValueError:  If the number of models does not equal the number of objectives
    """
    if not isinstance(models, list):
        raise ValueError("You must provide a list of models. One model per objective")
    if len(models) != ndim:
        raise ValueError("You must provide a list of models. One model per objective")


def validate_gpy_model(models: Any):
    """Make sure that all elements of the list a GPRegression models"""
    import GPy  # pylint:disable=import-outside-toplevel

    for model in models:
        if not isinstance(model, GPy.models.GPRegression):
            raise ValueError("The models must be an instance of GPy.model")


def validate_coregionalized_gpy(models: Any):
    """Make sure that model is a coregionalized GPR model"""
    if not isinstance(models, list):
        raise ValueError("You must provide a list of models with one element")

    from ..models.coregionalized import (  # pylint:disable=import-outside-toplevel
        GPCoregionalizedRegression,
    )

    if not isinstance(models[0], GPCoregionalizedRegression):
        raise ValueError(
            "Model must be a GPCoregionalized regression object from this package!"
        )


def validate_njobs(njobs: Any) -> int:
    """Make sure that njobs is an int > 1"""
    if not isinstance(njobs, int):
        raise ValueError("njobs musst be of type int")
    if njobs < 1:
        raise ValueError("njobs must be a number greater equal 1")
    return njobs


def validate_coef_var(coef_var: Any):
    """Make sure that the coef_var makes sense"""
    if not isinstance(coef_var, (float, int)):
        raise ValueError("coef_var must be of type float or int")
    if coef_var <= 0:
        raise ValueError("coef_var must be greater 0")

    return coef_var


def _validate_sklearn_gpr_model(model: Any) -> GaussianProcessRegressor:
    """Make sure that we deal with a GaussianProcessRegressor instance,
    if it is a fitted random or grid search instance, extract the model"""
    if isinstance(model, (RandomizedSearchCV, GridSearchCV)):
        try:
            if isinstance(model.best_estimator_, GaussianProcessRegressor):
                return model.best_estimator_

            raise ValueError(
                """If you provide a grid or random search instance,
it needs to contain a GaussianProcessRegressor instance."""
            )
        except AttributeError as not_fitted_exception:
            raise ValueError(
                "If you provide a grid or random search instance it needs to be fitted."
            ) from not_fitted_exception
    elif isinstance(model, GaussianProcessRegressor):
        return model
    raise ValueError("You need to provide a GaussianProcessRegressor instance.")


def validate_sklearn_gpr_models(
    models: Any, ndim: int
) -> List[GaussianProcessRegressor]:
    """Make sure that there is a list of GPR models, one model per objective"""
    validate_number_models(models, ndim)
    models_validated = []

    for model in models:
        models_validated.append(_validate_sklearn_gpr_model(model))

    return models_validated


def _validate_quantile_loss(lightgbmregressor):
    try:
        alpha = lightgbmregressor.alpha
        loss = lightgbmregressor.objective
    except AttributeError as missing_attribute:
        raise ValueError(
            """Make sure that you initialize at
least the first and last model with quantile loss.
"""
        ) from missing_attribute
    if loss != "quantile":
        raise ValueError(
            """Make sure that you initialize at
least the first and last model with quantile loss.
"""
        )
    assert alpha > 0


def validate_gbdt_models(models: Any, ndim: int) -> List[Iterable]:
    """Make sure that the number of iterables is equal to the number of objectives
    and that every iterable contains three LGBMRegressors.
    Also, we check that at least the first and last models use quantile loss"""

    validate_number_models(models, ndim)
    from lightgbm import LGBMRegressor  # pylint:disable=import-outside-toplevel

    for model_tuple in models:
        if len(model_tuple) != 3:
            raise ValueError(
                """The model list must contain
tuples with three LGBMRegressor instances.
        """
            )

        for counter, model in enumerate(model_tuple):
            if not isinstance(model, LGBMRegressor):
                raise ValueError(
                    """The model list must contain
tuples with three LGBMRegressor instances.
        """
                )

            if counter != 1:
                _validate_quantile_loss(model)

    return models


def validate_interquartile_scaler(interquartile_scaler: Any) -> float:
    """Make sure that the interquartile_scaler makes sense"""
    if not isinstance(interquartile_scaler, (float, int)):
        raise ValueError("interquartile_scaler must be a number.")

    if interquartile_scaler < 0:
        raise ValueError("interquartile_scaler must be a number greater 0.")

    return interquartile_scaler


def _is_jaxoptimizer(optimizer: Any) -> bool:
    from ..models.nt import JaxOptimizer  # pylint:disable=import-outside-toplevel

    return isinstance(optimizer, JaxOptimizer)


def validate_optimizers(optimizers: Any, ndim: int) -> Sequence:
    """Make sure that we can work with a Sequence if JaxOptimizer"""
    if not isinstance(optimizers, collections.Sequence):
        if not _is_jaxoptimizer(optimizers):
            raise ValueError(
                "You need to provide a `pyepal.models.nt.JaxOptimizer` instance"
            )
        return [deepcopy(optimizers) for _ in range(ndim)]

    if not len(optimizers) == ndim:
        raise ValueError(
            "If you provide a sequence it must have one optimizer per objective."
        )
    for optimizer in optimizers:
        if not _is_jaxoptimizer(optimizer):
            raise ValueError(
                "You need to provide a `pyepal.models.nt.JaxOptimizer` instance"
            )
    return optimizers


def validate_nt_models(models: Any, ndim: int) -> Sequence:
    """Make sure that we can work with a sequence of
    :py:func:`pyepal.pal.models.nt.NTModel`"""
    from pyepal.models.nt import NTModel  # pylint:disable=import-outside-toplevel

    if not isinstance(models, collections.Sequence):
        raise ValueError(
            "You need to provide a sequence of `pyepal.models.nt.NTModel` instances"
        )

    for model in models:
        if not len(models) == ndim:
            raise ValueError("You need to provide one model per objective.")
        if not isinstance(model, NTModel):
            raise ValueError(
                "You need to provide a sequence of `pyepal.models.nt.NTModel` instances"
            )

    return models


def validate_positive_integer_list(
    seq: Any, ndim: int, parameter_name: str = "Parameter"
) -> Sequence[int]:
    """Can be used, e.g., to validate and standardize the ensemble size
    and epochs input"""

    if not isinstance(seq, collections.Sequence):
        if (not isinstance(seq, int)) or (seq < 1):
            raise ValueError("{} must be a positive integer".format(parameter_name))
        return [seq] * ndim

    if not len(seq) == ndim:
        raise ValueError(
            "If you provide a sequence for {} its length must match \
                the number of objectives".format(
                parameter_name
            )
        )
    for elem in seq:
        if (not isinstance(elem, int)) or (elem < 1):
            raise ValueError("{} must be a positive integer".format(parameter_name))

    return seq


def validate_ranges(ranges: Any, ndim: int) -> Union[None, np.ndarray]:
    """Make sure that it has the correct numnber of elements and that all
    elements are positive."""
    if not isinstance(ranges, (np.ndarray, list)):
        return None

    if not len(ranges) == ndim:
        raise ValueError(
            "The number of elements in ranges must match the number of objectives."
        )
    for elem in ranges:
        if not elem > 0:
            raise ValueError("Ranges must be positive.")

    return np.array(ranges)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Run PAL with the same models for finite ensemble models
and infinite width models (`PALNT`)
"""

from typing import Sequence

import numpy as np
from sklearn.preprocessing import StandardScaler

from ..models.nt import JaxOptimizer, NTModel
from .pal_base import PALBase
from .validate_inputs import (
    validate_nt_models,
    validate_optimizers,
    validate_positive_integer_list,
)


# Again, the idea of having the core as pure functions outside of the class is that
# we could parallelize it easier in this way
def _ensemble_train_one_finite_width(  # pylint:disable=too-many-arguments, too-many-locals
    i: int,
    models: Sequence[NTModel],
    design_space: np.ndarray,
    objectives: np.ndarray,
    sampled: np.ndarray,
    optimizers: Sequence[JaxOptimizer],
    key: object,
    training_steps: Sequence[int],
    ensemble_size: Sequence[int],
):
    from jax import random  # pylint:disable=import-outside-toplevel
    from jax.api import grad, jit, vmap  # pylint:disable=import-outside-toplevel

    model = models[i]
    optimizer = optimizers[i]
    loss = jit(lambda params, x, y: 0.5 * np.mean((model.apply_fn(params, x) - y) ** 2))
    grad_loss = jit(lambda state, x, y: grad(loss)(optimizer.get_params(state), x, y))

    x_train = design_space[sampled[:, i]]

    scaler = StandardScaler()
    y_train = scaler.fit_transform(objectives[sampled[:, i], i].reshape(-1, 1))

    def train_network(key):
        _, params = model.init_fn(key, (-1, x_train.shape[1]))
        opt_state = optimizer.opt_init(params)

        for j in range(training_steps[i]):
            opt_state = optimizer.opt_update(
                j, grad_loss(opt_state, x_train, y_train), opt_state
            )

        return optimizer.get_params(opt_state)

    ensemble_key = random.split(key, ensemble_size[i])
    params = vmap(train_network)(ensemble_key)

    return params, scaler


def _ensemble_predict_one_finite_width(i: int, models: Sequence[NTModel], design_space):
    from jax.api import vmap  # pylint:disable=import-outside-toplevel

    model = models[i]

    ensemble_func = vmap(model.apply_fn, (0, None))(model.params, design_space)

    mean_func = np.reshape(np.mean(ensemble_func, axis=0), (-1,))
    std_func = np.reshape(np.std(ensemble_func, axis=0), (-1,))

    return mean_func, std_func


__all__ = ["PALJaxEnsemble", "NTModel", "JaxOptimizer"]


class PALJaxEnsemble(PALBase):  # pylint:disable=too-many-instance-attributes
    """Use PAL with and ensemble of finite-width neural networks.
    Note that we current assume that there is one model per output,
    i.e., we did not yet implement multihead support.
    """

    def __init__(self, *args, **kwargs):
        """Construct the PALJaxEnsemble instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (Sequence[NTModel]): You need to provide a sequence of
                 NTModel (`pyepal.models.nt.NTModel`).
                The elements of this dataclass are the `apply_fn`, `init_fn`,
                `kernel_fn` and `predict_fn` (for latter you can typically
                provide `None`).
                Can be constructed with
                :py:func:`pyepal.pal.models.nt.build_dense_network`.
            optimizer (Union[JaxOptimizer, Sequence[JaxOptimizer]]):
                Sequence of dataclasses with functions for a JAX optimizer,
                can be constructed with :py:func:`pyepal.pal.models.nt.get_optimizer`.
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            key (int): Seed to generate the key for the JAX
                pseudo-random number generator. Defaults to 10.
            training_steps (Union[int, Sequence[int]]): Number of epochs,
                the networks are trained. Defaults to 500.
            ensemble_size (Union[int, Sequence[int]]): Size of the ensemble, i.e.,
                over how many randomly initialized neural networks we average
                to obtain estimates of mean and standard deviation.
                Automatically vectorized using `vmap`.
                Defaults to 100.
        """
        from jax import random  # pylint:disable=import-outside-toplevel

        self.optimizers = validate_optimizers(
            kwargs.pop("optimizers"), kwargs.get("ndim")
        )

        self.training_steps = validate_positive_integer_list(
            kwargs.pop("training_steps", 500), kwargs.get("ndim")
        )
        self.ensemble_size = validate_positive_integer_list(
            kwargs.pop("ensemble_size", 100), kwargs.get("ndim")
        )
        self.key = random.PRNGKey(kwargs.pop("key", 10))
        self.design_space_scaler = StandardScaler()
        super().__init__(*args, **kwargs)
        self.models = validate_nt_models(self.models, self.ndim)

    def _set_data(self):
        self.design_space = self.design_space_scaler.fit_transform(self.design_space)

    def _train(self):
        for i in range(len(self.models)):
            params, scaler = _ensemble_train_one_finite_width(
                i,
                self.models,
                self.design_space,
                self.y,
                self.sampled,
                self.optimizers,
                self.key,
                self.training_steps,
                self.ensemble_size,
            )
            self.models[i].params = params
            self.models[i].scaler = scaler
            self.y[:, i] = scaler.transform(self.y[:, i].reshape(-1, 1)).flatten()

    def _predict(self):
        means, stds = [], []
        for i in range(len(self.models)):
            mean, std = _ensemble_predict_one_finite_width(
                i, self.models, self.design_space
            )
            means.append(mean.reshape(-1, 1))
            stds.append(std.reshape(-1, 1))

        self.means = np.hstack(means)
        self.std = np.hstack(stds)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=invalid-name
"""
Wrappers for Gaussian Process Regression models.

We typically use the GPy package as it offers most flexibility
for Gaussian processes in Python.
Typically, we use automatic relevance determination (ARD),
where one lengthscale parameter per input dimension is used.

If your task requires training on larger training sets,
you might consider replacing the models with their sparse
version but for the epsilon-PAL algorithm this typically shouldn't
be needed.

For kernel selection, you can have a look at
https://www.cs.toronto.edu/~duvenaud/cookbook/
Matérn, RBF and RationalQuadrat are good quick and dirty solutions
but have their caveats
"""

from typing import Tuple

import GPy
import numpy as np

from .coregionalized import GPCoregionalizedRegression

__all__ = [
    "get_matern_32_kernel",
    "get_matern_52_kernel",
    "get_ratquad_kernel",
    "build_coregionalized_model",
    "build_model",
    "predict",
    "predict_coregionalized",
    "set_xy_coregionalized",
]


def get_matern_32_kernel(NFEAT: int, ARD=True, **kwargs) -> GPy.kern.Matern32:
    """Matern-3/2 kernel without ARD"""
    return GPy.kern.Matern32(NFEAT, ARD=ARD, **kwargs)


def get_matern_52_kernel(NFEAT: int, ARD=True, **kwargs) -> GPy.kern.Matern52:
    """Matern-5/2 kernel without ARD"""
    return GPy.kern.Matern52(NFEAT, ARD=ARD, **kwargs)


def get_ratquad_kernel(NFEAT: int, ARD=True, **kwargs) -> GPy.kern.RatQuad:
    """Rational quadratic kernel without ARD"""
    return GPy.kern.RatQuad(NFEAT, ARD=ARD, **kwargs)


def build_coregionalized_model(
    X_train: np.array, y_train: np.array, kernel=None, w_rank: int = 1, **kwargs
) -> GPy.models.GPCoregionalizedRegression:
    """Wrapper for building a coregionalized GPR, it will have as many
    outputs as y_train.shape[1].
    Each output will have its own noise term"""
    NFEAT = X_train.shape[1]
    num_targets = y_train.shape[1]
    if isinstance(kernel, GPy.kern.src.kern.Kern):
        K = kernel
    else:
        K = get_matern_52_kernel(NFEAT)
    icm = GPy.util.multioutput.ICM(
        input_dim=NFEAT, num_outputs=num_targets, kernel=K, W_rank=w_rank
    )

    target_list = [y_train[:, i].reshape(-1, 1) for i in range(num_targets)]
    m = GPCoregionalizedRegression(
        [X_train] * num_targets, target_list, kernel=icm, normalizer=True, **kwargs
    )
    # We constrain the variance of the RBF/Matern ..
    # as the variance is now encoded in the kappa B of the ICM
    # Not constraining it would lead to a degeneracy
    m[".*ICM.*.variance"].constrain_fixed(1.0)
    # initialize the noise model
    m[".*Gaussian_noise_*"] = 0.1
    return m


def build_model(
    X_train: np.array, y_train: np.array, index: int = 0, kernel=None, **kwargs
) -> GPy.models.GPRegression:
    """Build a single-output GPR model"""
    NFEAT = X_train.shape[1]
    if isinstance(kernel, GPy.kern.src.kern.Kern):
        K = kernel
    else:
        K = get_matern_52_kernel(NFEAT)
    m = GPy.models.GPRegression(
        X_train, y_train[:, index].reshape(-1, 1), kernel=K, normalizer=True, **kwargs
    )
    m[".*Gaussian_noise_*"] = 0.1
    return m


def predict(model: GPy.models.GPRegression, X: np.array) -> Tuple[np.array, np.array]:
    """Wrapper function for the prediction method of a GPy regression model.
    It return the standard deviation instead of the variance"""
    assert isinstance(
        model, GPy.models.GPRegression
    ), "This wrapper function is written for GPy.models.GPRegression"
    mu, var = model.predict(X)
    return mu, np.sqrt(var)


def predict_coregionalized(
    model: GPy.models.GPCoregionalizedRegression, X: np.array, index: int = 0
) -> Tuple[np.array, np.array]:
    """Wrapper function for the prediction method of a coregionalized
    GPy regression model.
    It return the standard deviation instead of the variance"""
    assert isinstance(
        model, (GPy.models.GPCoregionalizedRegression, GPCoregionalizedRegression)
    ), "This wrapper function is written for GPy.models.GPCoregionalizedRegression"
    newX = np.hstack([X, index * np.ones_like(X)])
    mu_c0, var_c0 = model.predict(
        newX,
        Y_metadata={"output_index": index * np.ones((newX.shape[0], 1)).astype(int)},
    )

    return mu_c0, np.sqrt(var_c0)


def set_xy_coregionalized(model, X, y, mask=None):
    """Wrapper to update a coregionalized model with new data"""
    num_target = y.shape[1]
    if mask is None:
        X_array = [X] * num_target
        y_array = [y[:, i].reshape(-1, 1) for i in range(num_target)]

    else:
        X_array = [X[mask[:, i]] for i in range(num_target)]
        y_array = [y[mask[:, i], i].reshape(-1, 1) for i in range(num_target)]

    model.set_XY(X_array, y_array)

    return model



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Plotting utilities"""

from collections import defaultdict
from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import KFold

from .. import PALBase

plt.rcParams["font.family"] = "sans-serif"


__all__ = [
    "plot_bar_iterations",
    "plot_pareto_front_2d",
    "plot_residuals",
    "plot_histogram",
    "plot_jointplot",
]


def plot_bar_iterations(  # pylint:disable=invalid-name
    pareto_optimal: np.ndarray,
    non_pareto_points: np.ndarray,
    unclassified_points: np.ndarray,
    ax=None,
):
    """Plot stacked barplots for every step of the iteration.

    Args:
        pareto_optimal (np.ndarray): Number of pareto optimal points
            for every iteration.
        non_pareto_points (np.ndarray): Number of discarded points
            for every iteration
        unclassified_points (np.ndarray): Number of unclassified points
            for every iteration

    Returns:
        axis: matplotlib axis (the same that was provided as input
            or one from a new figure if no axis was provided)
    """
    assert (
        len(pareto_optimal) == len(non_pareto_points) == len(unclassified_points)
    ), "Make sure that the arrays have the same length"

    # We need numpy arrays as we assume that we can add the arrays
    # ToDo: We could potentially cast any iterable
    assert isinstance(pareto_optimal, np.ndarray), "The arguments must be numpy arrays"
    assert isinstance(
        non_pareto_points, np.ndarray
    ), "The arguments must be numpy arrays"
    assert isinstance(
        unclassified_points, np.ndarray
    ), "The arguments must be numpy arrays"

    if ax is None:
        _, ax = plt.subplots(1, 1)
    ax.bar(range(len(pareto_optimal)), unclassified_points, label="unclassified")
    ax.bar(
        range(len(pareto_optimal)),
        non_pareto_points,
        bottom=unclassified_points,
        label="discarded",
    )
    ax.bar(
        range(len(pareto_optimal)),
        pareto_optimal,
        bottom=np.array(non_pareto_points) + np.array(unclassified_points),
        label="Pareto optimal",
    )
    ax.set_xlabel("iteration (after initialization)")
    ax.set_ylabel("number of design points")

    return ax


def plot_pareto_front_2d(  # pylint:disable=too-many-arguments, invalid-name
    y_0: np.ndarray,
    y_1: np.ndarray,
    std_0: np.ndarray,
    std_1: np.ndarray,
    palinstance: PALBase,
    ax=None,
):
    """Plot a 2D pareto front, with the different categories
    indicated in color.

    Args:
        y_0 (np.ndarray): objective 0
        y_1 (np.ndarray): objective 1
        std_0 (np.ndarray): standard deviation objective 0
        std_1 (np.ndarray): standard deviation objective 0
        palinstance (PALBase): PAL instance
        ax (axix, optional): Matplotlib figure axis. Defaults to None.

    Returns:
        ax: matplotlib axis (the same that was provided as input
            or one from a new figure if no axis was provided)
    """
    # Here, we need numpy arrays for the indexing
    for array in [y_0, y_1, std_0, std_1]:
        assert isinstance(array, np.ndarray), "array must be a numpy array"
    assert (
        len(y_0)
        == len(y_1)
        == len(std_0)
        == len(std_1)
        == palinstance.number_design_points
    ), "Make sure that the arrays have the same length"

    if ax is None:
        _, ax = plt.subplots(1, 1)

    ax.errorbar(
        y_0,
        y_1,
        xerr=std_0,
        yerr=std_1,
        c="gray",
        alpha=0.3,
        label="all design points",
        fmt=".",
        capsize=5,
        zorder=0,
    )
    ax.scatter(
        y_0[palinstance.sampled_indices],
        y_1[palinstance.sampled_indices],
        c="blue",
        label="sampled",
        s=20,
        zorder=5,
    )
    ax.scatter(
        y_0[palinstance.discarded],
        y_1[palinstance.discarded],
        c="red",
        label="discarded",
        s=15,
        alpha=1,
        zorder=10,
    )
    ax.scatter(
        y_0[palinstance.pareto_optimal],
        y_1[palinstance.pareto_optimal],
        c="green",
        label="Pareto optimal",
        s=10,
        alpha=0.8,
        zorder=15,
    )

    return ax


def plot_histogram(
    y: np.ndarray, palinstance: PALBase, ax=None
):  # pylint:disable=invalid-name
    """Plot histograms, with maxima scaled to one
    and different categories indicated in color
    for one objective

    Args:
        y (np.ndarray): objective (measurement)
        palinstance (PALBase): instance of a PAL class
        ax (ax): Matplotlib figure axis

    Returns:
        ax: matplotlib axis (the same that was provided as input
            or one from a new figure if no axis was provided)ƒ
    """
    assert isinstance(y, np.ndarray), "Input array y must be a numpy array"
    assert y.ndim == 1, "Input array y must be one dimensional"
    assert (
        len(y) == palinstance.number_design_points
    ), "Length of y must equal the size of the design space"

    if ax is None:
        _, ax = plt.subplots(1, 1)
    heights, bins = np.histogram(y)
    bin_width = bins[1] - bins[0]
    ax.bar(
        bins[:-1],
        heights / heights.max(),
        bin_width,
        label="all design points",
        color="gray",
        alpha=0.6,
    )
    heights, bins = np.histogram(y[palinstance.sampled_indices])
    bin_width = bins[1] - bins[0]
    ax.bar(
        bins[:-1],
        heights / heights.max(),
        bin_width,
        label="sampled",
        color="blue",
        alpha=0.6,
    )

    heights, bins = np.histogram(y[palinstance.pareto_optimal])
    bin_width = bins[1] - bins[0]
    ax.bar(
        bins[:-1],
        heights / heights.max(),
        bin_width,
        label="Pareto optimal",
        color="green",
        alpha=0.6,
    )

    return ax


def plot_residuals(  # pylint:disable=invalid-name
    y: np.array,
    palinstance: PALBase,
    labels: Union[List[str], None] = None,
    figsize: tuple = (6.0, 4.0),
):
    """Plot signed residual (on y axis) vs fitted (on x axis)
    plot of sampled points.
    Will create suplots for y.ndim > 1.


    Args:
        y (np.array): Two-dimensional array with the objectives (measurements)
        palinstance (PALBase): "trained" PAL instance
        labels (Union[List[str], None], optional): Labels for each objective.
            Defaults to "objective [index]".
        figsize (tuple, optional): Figure size for each individual residual
            vs fitted objective plot. Defaults to (6.0, 4.0).

    Returns:
        fig: matplotlib Figure object
    """

    assert isinstance(y, np.ndarray), "Input array y must be a numpy array"
    assert (
        len(y) == palinstance.number_design_points
    ), "Length of y must equal the size of the design space"
    assert y.ndim == 2, "y must be a two-dimensional numpy array"
    assert (
        y.shape[1] == palinstance.ndim
    ), "y needs to be a two-dimensional array which \
        column number equals the number of targets"
    if palinstance.means is None:
        raise ValueError(
            "Predicted means is None. Execute run_one_step() \
                to obtain predicted means for each model."
        )

    num_targets = palinstance.ndim

    fig, ax = plt.subplots(  # pylint:disable=invalid-name
        num_targets,
        1,
        figsize=(figsize[0], num_targets * figsize[1]),
        tight_layout=True,
    )

    for index in range(num_targets):
        sampled = palinstance.sampled[:, index]
        fitted = palinstance.means[sampled, index]
        residuals = y[sampled, index] - fitted
        ax[index].scatter(fitted, residuals)
        ax[index].spines["top"].set_color("none")
        ax[index].spines["right"].set_color("none")
        ax[index].spines["left"].set_smart_bounds(True)
        ax[index].spines["bottom"].set_smart_bounds(True)

    if labels is None:
        labels = [f"objective {i}" for i in range(num_targets)]
    else:
        assert (
            len(labels) == num_targets
        ), "If labels are provided the length of the labels list \
            must equal the number of targets"

    for index in range(num_targets):
        ax[index].set_title(labels[index])

    return fig


def plot_jointplot(  # pylint:disable=invalid-name
    y: np.array,
    palinstance: PALBase,
    labels: Union[List[str], None] = None,
    figsize: tuple = (8.0, 6.0),
):
    """Plot a jointplot of the objective space with histograms on the diagonal
    and 2D-Pareto plots on the off-diagonal.

    Args:
        y (np.array): Two-dimensional array with the objectives (measurements)
        palinstance (PALBase): "trained" PAL instance
        labels (Union[List[str], None], optional): Labels for each objective.
            Defaults to "objective [index]".
        figsize (tuple, optional): Figure size for joint plot. Defaults to (8.0, 6.0).

    Returns:
        fig: matplotlib Figure object.
    """
    assert isinstance(y, np.ndarray), "Input array y must be a numpy array"
    assert (
        len(y) == palinstance.number_design_points
    ), "Length of y must equal the size of the design space"
    assert y.ndim == 2, "y must be a two-dimensional numpy array"
    assert (
        y.shape[1] == palinstance.ndim
    ), "y needs to be a two-dimensional array which column number \
        equals the number of targets"
    if (
        (palinstance.means is None)
        or (palinstance.std is None)
        or (palinstance.beta is None)
    ):
        raise ValueError(
            "Predicted means is None. Execute run_one_step() \
                to obtain predicted means for each model."
        )

    num_targets = y.shape[1]
    fig, ax = plt.subplots(  # pylint:disable=invalid-name
        num_targets, num_targets, figsize=figsize, tight_layout=True
    )

    for row in range(num_targets):
        for column in range(num_targets):
            if row == column:
                plot_histogram(y[:, row], palinstance, ax[row, column])
            else:
                plot_pareto_front_2d(
                    y[:, row],
                    y[:, column],
                    palinstance.std[:, row] * np.sqrt(palinstance.beta),
                    palinstance.std[:, column] * np.sqrt(palinstance.beta),
                    palinstance,
                    ax=ax[row, column],
                )

            ax[row, column].spines["top"].set_color("none")
            ax[row, column].spines["right"].set_color("none")
            ax[row, column].spines["left"].set_smart_bounds(True)
            ax[row, column].spines["bottom"].set_smart_bounds(True)

    if labels is None:
        labels = [f"objective {i}" for i in range(num_targets)]
    else:
        assert len(labels) == num_targets

    for index in range(num_targets):
        ax[index, 0].set_ylabel(labels[index])
        ax[num_targets - 1, index].set_xlabel(labels[index])

    ax[0, num_targets - 1].legend()

    return fig


def plot_learning_curve(  # pylint:disable=dangerous-default-value, too-many-arguments, too-many-locals
    palinstance,
    observations: np.ndarray,
    indices: np.ndarray = None,
    num_steps: int = 5,
    figsize: Tuple[float, float] = (8.0, 6.0),
    metrics: List[Tuple[str, callable]] = [
        ("MAE", mean_absolute_error),
        ("MSE", mean_squared_error),
        ("r2", r2_score),
    ],
    k_fold: object = KFold(n_splits=5, shuffle=True),
):
    """Plot learning curve

    Args:
        palinstance (pyepal.PALBase): A palinstance object
            (must have implemented _train and _predict methods)
        observations (np.ndarray): y array, measurements
        indices (np.ndarray, optional): Indices in the design space
            to wish the observations belong to. If none, it will default
            to the first ones. Defaults to None.
        num_steps (int, optional): Number of points on the learning curve.
            The maximum point will be the number of observation. Defaults to 5.
        figsize (Tuple[(float, float)], optional): Figure size.
            Defaults to (8.0, 6.0).
        metrics (List[Tuple[(str, callable)], optional):
            List of tuples (name, function) where the function is a metric function that
            takes an array of true values and predictions and returns a float.
            Defaults to [ ("MAE", mean_absolute_error), ("MSE", mean_squared_error),
            ("r2", r2_score), ].
        k_fold (object): Object with split method that takes an array and returns
            train/test indicies like the sklearn.model_selection.KFold object.
            Defaults to KFold(n_splits=5, shuffle=True).

    Returns:
        fig, dict: figure and dictionary with the learning curve results
    """
    if indices is None:
        indices = np.arange(0, len(observations))
    assert len(indices) == len(
        observations
    ), "The number of indices and observations must be equal"
    assert len(indices) > 5, "You need to use at least five points"
    grid = np.linspace(5, len(observations), num_steps, dtype=np.int8)
    grid = np.unique(grid)
    true_grid = []
    metrics_dict = defaultdict(list)

    for num_points in grid:
        sampled_indices = np.arange(0, num_points)
        sampled_indices = indices[sampled_indices]
        test_errors = defaultdict(list)
        num_training_points = None

        for train_idx, test_idx in k_fold.split(sampled_indices):
            train_idx = indices[train_idx]
            test_idx = [i for i in indices if i not in train_idx]

            num_training_points = len(train_idx)
            palinstance._reset()  # pylint: disable=protected-access
            palinstance.update_train_set(train_idx, observations[train_idx])
            palinstance._set_data()  # pylint: disable=protected-access
            palinstance._set_hyperparameters()  # pylint: disable=protected-access
            palinstance._train()  # pylint: disable=protected-access
            palinstance._predict()  # pylint: disable=protected-access

            for metric_name, metric_function in metrics:
                test_errors[metric_name].append(
                    metric_function(
                        observations[test_idx],
                        palinstance.means[test_idx],
                    )
                )
        for metric_name, results in test_errors.items():
            metrics_dict[metric_name].append(np.mean(results))
        true_grid.append(num_training_points)

    fig, ax = plt.subplots(  # pylint:disable=invalid-name
        len(metrics),
        1,
        figsize=figsize,
        tight_layout=True,
    )
    true_grid = list(true_grid)

    for i, items in enumerate(metrics_dict.items()):
        metric_name, metric_errors = items
        ax[i].plot(true_grid, metric_errors, label=metric_name)
        ax[i].set_ylabel(metric_name)

    ax[-1].set_xlabel("number of training points")
    metrics_dict["grid"] = grid
    return fig, metrics_dict



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""PAL using Sklearn GPR models"""
import concurrent.futures
from functools import partial

import numpy as np

from .pal_base import PALBase
from .validate_inputs import validate_njobs, validate_sklearn_gpr_models

__all__ = ["PALSklearn"]


def _train_model_picklable(i, models, design_space, objectives, sampled):
    model = models[i]
    model.fit(
        design_space[sampled[:, i]],
        objectives[sampled[:, i], i].reshape(-1, 1),
    )
    return model


class PALSklearn(PALBase):
    """PAL class for a list of Sklearn (GPR) models, with one model per objective"""

    def __init__(self, *args, **kwargs):
        """Construct the PALSklearn instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (list): Machine learning models. You can provide a list of
                GaussianProcessRegressor instances or a list of *fitted*
                RandomizedSearchCV/GridSearchCV instances with
                GaussianProcessRegressor models
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            n_jobs (int): Number of parallel processes that are used to fit
                the GPR models. Defaults to 1.
        """
        self.n_jobs = validate_njobs(kwargs.pop("n_jobs", 1))
        super().__init__(*args, **kwargs)

        self.models = validate_sklearn_gpr_models(self.models, self.ndim)

    def _set_data(self):
        pass

    def _train(self):
        train_single_partial = partial(
            _train_model_picklable,
            models=self.models,
            design_space=self.design_space,
            objectives=self.y,
            sampled=self.sampled,
        )
        models = []
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.n_jobs
        ) as executor:
            for model in executor.map(train_single_partial, range(self.ndim)):
                models.append(model)
        self.models = models

    def _predict(self):
        means, stds = [], []
        for model in self.models:
            mean, std = model.predict(self.design_space, return_std=True)
            means.append(mean.reshape(-1, 1))
            stds.append(std.reshape(-1, 1))

        self.means = np.hstack(means)
        self.std = np.hstack(stds)

    def _set_hyperparameters(self):
        pass



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""PAL for coregionalized GPR models"""

import numpy as np

from .pal_base import PALBase
from .schedules import linear

__all__ = ["PALCoregionalized"]


class PALCoregionalized(PALBase):
    """PAL class for a coregionalized GPR model"""

    def __init__(self, *args, **kwargs):
        """Construct the PALCoregionalized instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (list): Machine learning models
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            restarts (int): Number of random restarts that are used for hyperparameter
                optimization. Defaults to 20.
            parallel (bool): If true, model hyperparameters are optimized in parallel,
                using the GPy implementation. Defaults to False.
        """
        from .validate_inputs import (  # pylint:disable=import-outside-toplevel
            validate_coregionalized_gpy,
        )

        self.restarts = kwargs.pop("restarts", 20)
        self.parallel = kwargs.pop("parallel", False)
        assert isinstance(
            self.parallel, bool
        ), "the parallel keyword must be of type bool"
        assert isinstance(
            self.restarts, int
        ), "the restarts keyword must be of type int"
        super().__init__(*args, **kwargs)
        validate_coregionalized_gpy(self.models)

    def _set_data(self):
        from ..models.gpr import (  # pylint:disable=import-outside-toplevel
            set_xy_coregionalized,
        )

        self.models[0] = set_xy_coregionalized(
            self.models[0],
            self.design_space[self.sampled_indices],
            self.y[self.sampled_indices],
            self.sampled[self.sampled_indices],
        )

    def _train(self):
        pass

    def _predict(self):
        from ..models.gpr import (  # pylint:disable=import-outside-toplevel
            predict_coregionalized,
        )

        means, stds = [], []
        for i in range(self.ndim):
            mean, std = predict_coregionalized(self.models[0], self.design_space, i)
            means.append(mean.reshape(-1, 1))
            stds.append(std.reshape(-1, 1))

        self.means = np.hstack(means)
        self.std = np.hstack(stds)

    def _set_hyperparameters(self):
        self.models[0].optimize_restarts(self.restarts, parallel=self.parallel)

    def _should_optimize_hyperparameters(self) -> bool:
        return linear(self.iteration, 10)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some utilities to work with LightGBM models"""
from typing import List, Tuple, Union

import numpy as np
from lightgbm import LGBMRegressor

__all__ = ["build_gbdt_tuple"]


def build_gbdt_tuple(
    alphas: Union[List[float], np.ndarray, Tuple[float, float, float]] = [
        0.25,
        0.5,
        0.75,
    ],
    **kwargs  # pylint:disable=dangerous-default-value
) -> Tuple[LGBMRegressor, LGBMRegressor, LGBMRegressor]:
    """Build a Tuple of LGBMRegressors in the correct
    format for PALGBDT

    Args:
        alphas (Union[List[float], np.ndarray, Tuple[float]], optional):
            Quantile levels.
            Defaults to [0.16, 0.5, 0.84]
            (about 68% of the probability under a
            normal curve lies between 𝜇±𝜎.)

    Returns:
        Tuple[LGBMRegressor, LGBMRegressor, LGBMRegressor]
    """
    return (
        LGBMRegressor(objective="quantile", alpha=alphas[0], **kwargs),
        LGBMRegressor(objective="quantile", alpha=alphas[1], **kwargs),
        LGBMRegressor(objective="quantile", alpha=alphas[2], **kwargs),
    )



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to build neutral tangents models for PALNT

Depending on the dataset there might be some issues with these models,
some tricks are listed in https://github.com/google/neural-tangents/issues/76

1. Use Erf as activation
2. Initialize the weights with larger standard deviation
3. Standardize the data

The first two points are done by default in the `build_dense_network` function

Note that following the law of total variance the prior, intialized via
W_std and b_std give an upper bound on the std of the posterior
"""

from dataclasses import dataclass
from typing import Callable, Sequence, Union


@dataclass
class NTModel:
    """Defining a dataclass for neural tangents models"""

    # Initialization functions construct parameters for neural networks
    # given a random key and an input shape.
    init_fn: Callable
    # Apply functions do computations with finite-width neural networks.
    apply_fn: Callable
    kernel_fn: Callable
    predict_fn: Union[Callable, None] = None
    scaler: Union[Callable, None] = None  # Used to store Standard Scaler objects
    params: Union[list, None] = None  # Used to store parameters for the ensemble models


@dataclass
class JaxOptimizer:
    """Defining a dataclass for a JAX optimizer"""

    opt_init: Callable
    opt_update: Callable
    get_params: Callable


__all__ = ["NTModel", "build_dense_network", "JaxOptimizer"]


def build_dense_network(
    hidden_layers: Sequence[int],
    activations: Union[Sequence, str] = "erf",
    w_std: float = 2.5,
    b_std=1,
) -> NTModel:
    """Utility function to build a simple feedforward network with the
    neural tangents library.

    Args:
        hidden_layers (Sequence[int]): Iterable with the number of neurons.
            For example, [512, 512]
        activations (Union[Sequence, str], optional):
            Iterable with neural_tangents.stax axtivations or "relu" or "erf".
            Defaults to "erf".
        w_std (float): Standard deviation of the weight distribution.
        b_std (float): Standard deviation of the bias distribution.

    Returns:
        NTModel: jiited init, apply and
            kernel functions, predict_function (None)
    """
    from jax.config import config  # pylint:disable=import-outside-toplevel

    config.update("jax_enable_x64", True)
    from jax import jit  # pylint:disable=import-outside-toplevel
    from neural_tangents import stax  # pylint:disable=import-outside-toplevel

    assert len(hidden_layers) >= 1, "You must provide at least one hidden layer"
    if activations is None:
        activations = [stax.Relu() for _ in hidden_layers]
    elif isinstance(activations, str):
        if activations.lower() == "relu":
            activations = [stax.Relu() for _ in hidden_layers]
        elif activations.lower() == "erf":
            activations = [stax.Erf() for _ in hidden_layers]
    else:
        for activation in activations:
            assert callable(
                activation
            ), "You need to provide `neural_tangents.stax` activations"

    assert len(activations) == len(
        hidden_layers
    ), "The number of hidden layers should match the number of nonlinearities"
    stack = []

    for hidden_layer, activation in zip(hidden_layers, activations):
        stack.append(stax.Dense(hidden_layer, W_std=w_std, b_std=b_std))
        stack.append(activation)

    stack.append(stax.Dense(1, W_std=w_std, b_std=b_std))

    init_fn, apply_fn, kernel_fn = stax.serial(*stack)

    return NTModel(init_fn, jit(apply_fn), jit(kernel_fn, static_argnums=(2,)), None)


def get_optimizer(
    learning_rate: float = 1e-4, optimizer="sdg", optimizer_kwargs: dict = None
) -> JaxOptimizer:
    """Return a `JaxOptimizer` dataclass for a JAX optimizer

    Args:
        learning_rate (float, optional): Step size. Defaults to 1e-4.
        optimizer (str, optional): Optimizer type (Allowed types: "adam",
            "adamax", "adagrad", "rmsprop", "sdg"). Defaults to "sdg".
        optimizer_kwargs (dict, optional): Additional keyword arguments
            that are passed to the optimizer. Defaults to None.

    Returns:
        JaxOptimizer
    """
    from jax.config import config  # pylint:disable=import-outside-toplevel

    config.update("jax_enable_x64", True)
    from jax import jit  # pylint:disable=import-outside-toplevel
    from jax.experimental import optimizers  # pylint:disable=import-outside-toplevel

    if optimizer_kwargs is None:
        optimizer_kwargs = {}
    optimizer = optimizer.lower()
    if optimizer == "adam":
        opt_init, opt_update, get_params = optimizers.adam(
            learning_rate, **optimizer_kwargs
        )
    elif optimizer == "adagrad":
        opt_init, opt_update, get_params = optimizers.adagrad(
            learning_rate, **optimizer_kwargs
        )
    elif optimizer == "adamax":
        opt_init, opt_update, get_params = optimizers.adamax(
            learning_rate, **optimizer_kwargs
        )
    elif optimizer == "rmsprop":
        opt_init, opt_update, get_params = optimizers.rmsprop(
            learning_rate, **optimizer_kwargs
        )
    else:
        opt_init, opt_update, get_params = optimizers.sgd(
            learning_rate, **optimizer_kwargs
        )

    opt_update = jit(opt_update)

    return JaxOptimizer(opt_init, opt_update, get_params)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Use PAL with the neural tangent library.
This allows to perform
1. Exact Bayesian inference (NNGP)
2. Inference using gradient descent with MSE loss (NTK)

Note that the neural tangent code usually assumes mean-zero Gaussians


Reducing predict_fn_kwargs['diag_reg'] typically improves the interpolation
quality
"""

from typing import Sequence, Tuple

import numpy as np
from sklearn.preprocessing import StandardScaler

from ..models.nt import NTModel
from .pal_base import PALBase
from .validate_inputs import validate_nt_models

__all__ = ["PALNT", "NTModel"]

# We move those functions out of the class so that we can parallelize them
def _set_one_infinite_width_model(  # pylint:disable=too-many-arguments
    i: int,
    models: Sequence[NTModel],
    design_space: np.ndarray,
    objectives: np.ndarray,
    sampled: np.ndarray,
    predict_fn_kwargs: dict = None,
) -> Tuple[callable, StandardScaler]:
    from jax.config import config  # pylint:disable=import-outside-toplevel

    config.update("jax_enable_x64", True)
    import neural_tangents as nt  # pylint:disable=import-outside-toplevel

    if predict_fn_kwargs is None:
        predict_fn_kwargs = {"diag_reg": 1e-3}
    model = models[i]
    kernel_fn = model.kernel_fn
    scaler = StandardScaler()
    y = scaler.fit_transform(  # pylint:disable=invalid-name
        objectives[sampled[:, i], i].reshape(-1, 1)
    )
    predict_fn = nt.predict.gradient_descent_mse_ensemble(
        kernel_fn,
        design_space[sampled[:, i]],
        y,
        **predict_fn_kwargs,
    )

    return predict_fn, scaler


def _predict_one_infinite_width_model(
    i: int, models: Sequence[NTModel], design_space: np.ndarray, kernel: str
):
    predict_fn = models[i].predict_fn
    mean, covariance = predict_fn(  # type: ignore
        x_test=design_space,
        get=kernel,
        compute_cov=True,
    )

    return mean.flatten(), np.sqrt(np.diag(covariance))


class PALNT(PALBase):
    """ε-PAL with neural tangents models
    (in the current implementation, one model per objective)"""

    def __init__(self, *args, **kwargs):
        """Construct the PALNT instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (Sequence[NTModel]): You need to provide a sequence of
                 NTModel (`pyepal.models.nt.NTModel`).
                The elements of this dataclass are the `apply_fn`, `init_fn`,
                `kernel_fn` and `predict_fn` (for latter you can typically
                provide `None`)
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            kernel (str, optional): The kernel type you want to use ('nngp' or 'ntk').
                NNGP corresponds to a Neural Network Gaussian Process, frist established
                by Neal in 1994.  NTK refers to the neural tangent kernel,
                i.e., the linear approximation of an infinite width neural network
                (LeCun initialized) trained with gradient descent (Jacot et al., 2018).
                Defaults to 'nngp'.
        """

        self.kernel = kwargs.pop("kernel", "nngp")
        self.design_space_scaler = StandardScaler()
        super().__init__(*args, **kwargs)
        self.models = validate_nt_models(self.models, self.ndim)

    def _set_data(self):
        self.design_space = self.design_space_scaler.fit_transform(self.design_space)
        for i in range(len(self.models)):

            predict_fn, scaler = _set_one_infinite_width_model(
                i,
                self.models,
                self.design_space,
                self.y,
                self.sampled,
            )
            self.models[i].predict_fn = predict_fn
            self.models[i].scaler = scaler
            self.y[:, i] = scaler.transform(self.y[:, i].reshape(-1, 1)).flatten()

    def _train(self):
        pass

    def _predict(self):
        means, stds = [], []
        for i in range(len(self.models)):
            mean, std = _predict_one_infinite_width_model(
                i,
                self.models,
                self.design_space,
                self.kernel,
            )
            means.append(mean.reshape(-1, 1))
            stds.append(std.reshape(-1, 1))

        self.means = np.hstack(means)
        self.std = np.hstack(stds)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provides some scheduling functions
that can be used to implement the _should_optimize_hyperparameters function"""
import math

import numpy as np

__all__ = ["linear", "exp_decay"]


def linear(iteration: int, frequency: int = 10) -> bool:
    """Optimize hyperparameters at equally spaced intervals

    Args:
        iteration (int): current iteration
        frequency (int, optional): Spacing between the True outputs. Defaults to 10.

    Returns:
        bool: True if iteration can be divided by frequency without remainder
    """
    if iteration == 1:
        return True
    if iteration % frequency == 0:
        return True
    return False


def exp_decay(iteration: int, base: int = 10) -> bool:
    """Optimize hyperparameters at logartihmically spaced intervals

    Args:
        iteration (int): current iteration
        base (int, optional): Base of the logarithm. Defaults to 10.

    Returns:
        bool: True if iteration is on the log scaled grid
    """
    if iteration == 1:
        return True

    result = math.log(iteration, base)
    if np.abs(result - round(result)) < 0.00001:
        return True
    return False



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""PyePAL"""
from ._version import get_versions
from .models.nt import JaxOptimizer, NTModel
from .pal.pal_base import PALBase
from .pal.pal_coregionalized import PALCoregionalized
from .pal.pal_finite_ensemble import PALJaxEnsemble
from .pal.pal_gbdt import PALGBDT
from .pal.pal_gpy import PALGPy
from .pal.pal_neural_tangent import PALNT
from .pal.pal_sklearn import PALSklearn
from .pal.utils import (
    exhaust_loop,
    get_hypervolume,
    get_kmeans_samples,
    get_maxmin_samples,
)

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "PALBase",
    "PALCoregionalized",
    "PALGBDT",
    "PALGPy",
    "PALSklearn",
    "PALJaxEnsemble",
    "PALNT",
    "NTModel",
    "JaxOptimizer",
    "exhaust_loop",
    "get_hypervolume",
    "get_kmeans_samples",
    "get_maxmin_samples",
]



# -*- coding: utf-8 -*-
# pylint:disable=anomalous-backslash-in-string
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Base class for PAL"""

import logging
import warnings
from copy import deepcopy
from typing import List, Union

import numpy as np
from sklearn.metrics import mean_absolute_error

from .core import (
    _get_max_wt,
    _get_max_wt_all,
    _get_uncertainty_regions,
    _pareto_classify,
    _uncertainty,
    _union,
)
from .validate_inputs import (
    base_validate_models,
    validate_beta_scale,
    validate_coef_var,
    validate_delta,
    validate_epsilon,
    validate_goals,
    validate_ndim,
    validate_ranges,
)

PAL_LOGGER = logging.getLogger("PALLogger")
PAL_LOGGER.setLevel(logging.INFO)
CONSOLE_HANDLER = logging.StreamHandler()
CONSOLE_FORMAT = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
CONSOLE_HANDLER.setFormatter(CONSOLE_FORMAT)
PAL_LOGGER.addHandler(CONSOLE_HANDLER)


__all__ = ["PALBase", "PAL_LOGGER"]


class PALBase:  # pylint:disable=too-many-instance-attributes, too-many-public-methods
    """PAL base class"""

    def __init__(  # pylint:disable=too-many-arguments
        self,
        X_design: np.array,
        models: list,
        ndim: int,
        epsilon: Union[list, float] = 0.01,
        delta: float = 0.05,
        beta_scale: float = 1 / 9,
        goals: List[str] = None,
        coef_var_threshold: float = 3,
        ranges: Union[np.ndarray, None] = None,
    ):
        """Initialize the PAL instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (list): Machine learning models
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            ranges (np.ndarray, optional): Numpy array of length ndmin,
                where each element contains the value range of given objective.
                If this is provided, we will use :math:`\epsilon \cdot ranges`
                to computer the uncertainties of the hyperrectangles instead
                of the default behavior :math:`\epsilon \cdot |\mu|`

        """
        self.cross_val_points = 10  # maybe we make it an argument at some point
        self.ndim = validate_ndim(ndim)
        self.epsilon = validate_epsilon(epsilon, self.ndim)
        self.delta = validate_delta(delta)
        self.beta_scale = validate_beta_scale(beta_scale)
        self.pareto_optimal = np.array([False] * len(X_design))
        self.discarded = np.array([False] * len(X_design))
        self.sampled = np.array([[False] * self.ndim] * len(X_design))
        self.unclassified = np.array([True] * len(X_design))
        self.rectangle_ups: np.array = None
        self.rectangle_lows: np.array = None
        self.models = base_validate_models(models)
        self.iteration = 1
        design_space_size = len(X_design)
        self.coef_var_threshold = validate_coef_var(coef_var_threshold)
        self.coef_var_mask = np.array([True] * design_space_size)
        # means/std are the model predictions
        self.means: np.array = None
        self.std: np.array = None

        self.design_space = X_design
        self.beta = None
        self.goals = validate_goals(goals, ndim)
        self.ranges = validate_ranges(ranges, ndim)

        # self.y is what needs to be used for train/predict
        # as there the data has been turned into maximization
        # self._y contains the data as provided by the user
        self.y = np.zeros((design_space_size, self.ndim))  # pylint:disable=invalid-name
        self._y = self.y
        # measurement_uncertainity is provided in update_train_set by the user
        self.measurement_uncertainty = np.zeros((design_space_size, self.ndim))
        self._has_train_set = False

    def __repr__(self):
        return f"pyepal at iteration {self.iteration}. \
        {self.number_pareto_optimal_points} Pareto optimal points, \
        {self.number_discarded_points} discarded points, \
        {self.number_unclassified_points} unclassified points."

    def _uses_fixed_epsilon(self):
        if self.ranges is not None:
            return True
        return False

    @property
    def uses_fixed_epsilon(self):
        """True if it uses the fixed epsilon :math:`\epsilon \cdot ranges`"""
        return self._uses_fixed_epsilon()

    def _reset(self):
        self.pareto_optimal = np.array([False] * self.number_design_points)
        self.discarded = np.array([False] * self.number_design_points)
        self.sampled = np.array([[False] * self.ndim] * self.number_design_points)
        self.unclassified = np.array([True] * self.number_design_points)
        self.rectangle_ups: np.array = None
        self.rectangle_lows: np.array = None

        self.iteration = 1

        self.coef_var_mask = np.array([True] * self.number_design_points)

        # means/std are the model predictions
        self.means: np.array = None
        self.std: np.array = None
        self.beta = None

        # self.y is what needs to be used for train/predict
        # as there the data has been turned into maximization
        # self._y contains the data as provided by the user
        self.y = np.zeros(
            (self.number_design_points, self.ndim)
        )  # pylint:disable=invalid-name
        self._y = self.y
        # measurement_uncertainity is provided in update_train_set by the user
        self.measurement_uncertainty = np.zeros((self.number_design_points, self.ndim))
        self._has_train_set = False

    @property
    def sampled_mask(self):
        """Create a mask for the sampled points
        We count a point as sampled if at least one objective has
        been measured, i.e., self.sampled is a N * number objectives
        array in which some columns can be false if a measurement
        has not been performed"""
        return self.sampled.sum(axis=1) > 0

    @property
    def number_design_points(self):
        """Return the number of points in the design space"""
        return len(self.design_space)

    @property
    def pareto_optimal_points(self):
        """Return the pareto optimal points"""
        return self.design_space[self.pareto_optimal]

    @property
    def sampled_points(self):
        """Return the sampled points"""
        return self.design_space[self.sampled_indices]

    @property
    def discarded_points(self):
        """Return the discarded points"""
        return self.design_space[self.discarded]

    @property
    def unclassified_points(self):
        """Return the discarded points"""
        return self.design_space[self.unclassified]

    @property
    def pareto_optimal_indices(self):
        """Return the indices of the Pareto optimal points"""
        return np.where(self.pareto_optimal)[0]

    @property
    def sampled_indices(self):
        """Return the indices of the sampled points"""
        return np.unique(np.where(self.sampled)[0])

    @property
    def discarded_indices(self):
        """Return the indices of the discarded points"""
        return np.where(self.discarded)[0]

    @property
    def unclassified_indices(self):
        """Return the indices of the unclassified points"""
        return np.where(self.unclassified)[0]

    @property
    def number_pareto_optimal_points(self):
        """Return the number of Pareto optimal points"""
        return sum(self.pareto_optimal)

    @property
    def number_discarded_points(self):
        """Return the nnumber of discarded points"""
        return sum(self.discarded)

    @property
    def number_unclassified_points(self):
        """Return the number of unclassified points"""
        return sum(self.unclassified)

    @property
    def number_sampled_points(self):
        """Return the number of sampled points"""
        return len(self.sampled_indices)

    @property
    def hyperrectangle_sizes(self):
        """Return the sizes of the hyperrectangles"""
        return _uncertainty(self.rectangle_ups, self.rectangle_lows, self.means)

    def _update_beta(self):
        """Update beta according to section 7.2. of the epsilon-PAL paper"""
        self.beta = (
            self.beta_scale
            * 2
            * np.log(
                self.ndim
                * self.number_design_points
                * np.square(np.pi)
                * np.square(self.iteration + 1)
                / (6 * self.delta)
            )
        )

    def _log(self):
        pass

    def _should_optimize_hyperparameters(self) -> bool:  # pylint:disable=no-self-use
        return True

    def _predict(self):
        raise NotImplementedError("The predict function is not implemented")

    def _set_hyperparameters(self):
        pass

    def _turn_to_maximization(self):
        self.y = self._y * self.goals

    def _train(self):
        pass

    def _set_data(self):
        pass

    def _crossvalidate(self):
        sampled_original = deepcopy(self.sampled)
        sampled_idx_original = self.sampled_indices
        errors = []
        # this step is to make the code not to complicate
        # and deal with both large and small sample sizes
        # in small samples sizes, we want to do leave-on-out CV
        # in large samples this is too expensive,
        # hence we chose a random subset, for which we
        # test the model

        sample_subset = np.random.choice(
            sampled_idx_original,
            min([self.cross_val_points, len(sampled_idx_original)]),
            replace=False,
        )
        for sampled_idx in sample_subset:
            # make sure that we do not run into errors due to np.nan
            if self.sampled[sampled_idx].sum() == self.ndim:
                # copy here is important, otherewise all
                # points we set to False remain False
                self.sampled = sampled_original.copy()
                self.sampled[sampled_idx, :] = False

                self._set_data()
                if self._should_optimize_hyperparameters():
                    self._set_hyperparameters()
                self._train()
                self._predict()
                error = mean_absolute_error(
                    self.y[sampled_idx], self.means[sampled_idx]
                )
                errors.append(error)

        self.sampled = sampled_original
        return np.array(errors).mean()

    def should_cross_validate(self):
        """Override for more complex cross validation schedules"""
        if (self.iteration == 1) & (self.cross_val_points > 0):
            return True
        return False

    def _update_hyperrectangles(self, new_indices: np.ndarray = None):
        """Computes new hyperrectangles based on beta,
        the means and the standard deviations.
        If the iteration is > 0,
        then it uses iterative intersection to ensure that the size of the
        hyperrectangles is decreasing.

        Args:
            new_indices (np.ndarray): If provided, it will not use the
                iterative intersection algorithm for these hyperrectangles.
                Instead, it will just use the scaled edges based on the
                model's prediction. Defaults to None.
        """
        lows, ups = _get_uncertainty_regions(self.means, self.std, np.sqrt(self.beta))
        if self.iteration == 1:
            # initialization
            self.rectangle_lows, self.rectangle_ups = lows, ups
        else:
            if new_indices is None:
                self.rectangle_lows, self.rectangle_ups = _union(
                    self.rectangle_lows, self.rectangle_ups, lows, ups
                )
            else:
                not_new = np.array(
                    [
                        i
                        for i in range(self.number_design_points)
                        if i not in new_indices
                    ]
                )
                self.rectangle_lows[new_indices] = lows[new_indices]
                self.rectangle_ups[new_indices] = ups[new_indices]
                self.rectangle_lows[not_new], self.rectangle_ups[not_new] = _union(
                    self.rectangle_lows[not_new],
                    self.rectangle_ups[not_new],
                    lows[not_new],
                    ups[not_new],
                )

    def _update_coef_var_mask(self):
        """Update the mask array of elements that have variance below
        the coefficient of variation threshold"""
        # small workaround to avoid potential bugs due to division by zero
        # this will fail if everything is zero in this case,
        # I feel we might just give up for now,
        # in the future we can think fo just applying a shift
        if self.means.sum() != 0:
            means_no_zero = self.means.copy()
            means_no_zero[means_no_zero == 0] = np.median(means_no_zero)
            self.coef_var_mask = (
                np.max(self.std / means_no_zero, axis=1) < self.coef_var_threshold
            )
        else:
            mean_variation = self.std.mean()
            self.coef_var_mask = (
                np.max(self.std / mean_variation, axis=1) < self.coef_var_threshold
            )

    def _classify(self):
        self._update_coef_var_mask()
        if self.uses_fixed_epsilon:
            pareto_optimal, discarded, unclassified = _pareto_classify(
                self.pareto_optimal[self.coef_var_mask],
                self.discarded[self.coef_var_mask],
                self.unclassified[self.coef_var_mask],
                self.rectangle_lows[self.coef_var_mask],
                self.rectangle_ups[self.coef_var_mask],
                self.epsilon * self.ranges,
                is_fixed_epsilon=True,
            )
        else:
            pareto_optimal, discarded, unclassified = _pareto_classify(
                self.pareto_optimal[self.coef_var_mask],
                self.discarded[self.coef_var_mask],
                self.unclassified[self.coef_var_mask],
                self.rectangle_lows[self.coef_var_mask],
                self.rectangle_ups[self.coef_var_mask],
                self.epsilon,
                is_fixed_epsilon=False,
            )
        self.pareto_optimal[self.coef_var_mask] = pareto_optimal
        self.discarded[self.coef_var_mask] = discarded
        self.unclassified[self.coef_var_mask] = unclassified

    def _replace_by_measurements(
        self, replace_mean: bool = True, replace_std: bool = True
    ):
        """Implements one "trick". Instead of using the GPR
        predictions for the sampled points we use the data that
        was actually measured and the actual uncertainty."""
        if replace_mean:
            self.means[self.sampled] = self.y[self.sampled]
        if replace_std:
            self.std[self.sampled] = self.measurement_uncertainty[self.sampled]

    def run_one_step(  # pylint:disable=too-many-arguments
        self,
        batch_size: int = 1,
        pooling_method: str = "fro",
        sample_discarded: bool = False,
        use_coef_var: bool = True,
        replace_mean: bool = True,
        replace_std: bool = True,
    ) -> Union[np.array, None]:
        """[summary]

        Args:
            batch_size (int, optional): Number of indices that will be returned.
                Defaults to 1.
            pooling_method (str): Method that is used to aggregate
                the uncertainty in different objectives into one scalar.
                Available options are:  "fro" (Frobenius/Euclidean norm), "mean",
                "median". Defaults to "fro".
            sample_discarded (bool): if true, it will sample from all points
                and not only from the unclassified and Pareto optimal ones
            use_coef_var (bool): If True, uses the coefficient of variation instead of
                the unscaled rectangle sizes
            replace_mean (bool): If true uses the measured means for the sampled points
            replace_std (bool): If true uses the measured standard deviation for the
                sampled points

        Raises:
            ValueError: In case the PAL instance was not initialized with
                measurements.

        Returns:
            Union[np.array, None]: Returns array of indices if there are
                unclassified points that can be sample left.
        """
        if not self._has_train_set:
            raise ValueError(
                "You need to provide a set of points to train the model on.\
               Before the first iteration, call the update_train_set() function."
            )

        self._set_data()
        if self.should_cross_validate():
            self._compare_mae_variance()

        if self._should_optimize_hyperparameters():
            self._set_hyperparameters()

        self._train()
        self._predict()

        self._update_beta()
        self._replace_by_measurements(
            replace_mean=replace_mean, replace_std=replace_std
        )
        self._update_hyperrectangles()
        self._classify()
        samples = np.array([], dtype=np.int)
        if sum((self.pareto_optimal | self.unclassified) & ~self.sampled_mask) == 0:
            PAL_LOGGER.info("No point that we could sample left.")
            return None
        if sum(self.unclassified):
            for _ in range(batch_size):
                sampled_idx = self.sample(
                    exclude_idx=samples,
                    pooling_method=pooling_method,
                    sample_discarded=sample_discarded,
                    use_coef_var=use_coef_var,
                )
                samples = np.append(samples, [sampled_idx])
                self._log()

            self._log()
            self.iteration += 1

            return samples
        PAL_LOGGER.info("Done. No unclassified point left.")
        return None

    def _compare_mae_variance(self):
        mae = np.nan_to_num(self._crossvalidate(), nan=np.inf)
        self._predict()
        mean_std = self.std.mean()
        if mae >= mean_std:
            warnings.warn(
                """The mean absolute error in crossvalidation is {:.2f},
the mean standard deviation is {:.2f}.
Your model might not be predictive and/or overconfident.
In the docs, you find hints on how to make models more robust.""".format(
                    mae, mean_std
                ),
                UserWarning,
            )
        else:
            PAL_LOGGER.info(
                """The mean absolute error in crossvalidation is
{:.2f}, the mean variance is {:.2f}.""".format(
                    mae, mean_std
                )
            )

    def update_train_set(
        self,
        indices: np.ndarray,
        measurements: np.ndarray,
        measurement_uncertainty: np.ndarray = None,
    ):
        """Update training set following a measurement

        Args:
            indices (np.ndarray): Indices of design space at which
                the measurements were taken
            measurements (np.ndarray): Measured values, 2D array.
                the length must equal the length of the indices array.
                the second direction must equal the number of objectives.
                If an objective is missing, provide np.nan. For example,
                np.array([1, 1, np.nan])
            measurement_uncertainty (np.ndarray): uncertainty in the measuremens,
                if not provided (None) will be zero. If it is not None, it must be
                an array with the same shape as the measurements
                If an objective is missing, provide np.nan.
                For example, np.array([1, 1, np.nan])
        """
        self._has_train_set = True
        assert measurements.shape[1] == self.ndim
        assert len(indices) == len(measurements)
        if measurement_uncertainty is not None:
            assert measurement_uncertainty.shape == measurements.shape
        else:
            measurement_uncertainty = np.zeros(measurements.shape)
        self._y[indices] = measurements
        self.measurement_uncertainty[indices] = measurement_uncertainty
        # This sets "sampled" to False for the objectives that have not been measured
        self.sampled[indices] = ~np.isnan(measurements)
        self._turn_to_maximization()

    def augment_design_space(  # pylint: disable=invalid-name
        self, X_design: np.ndarray, classify: bool = False, clean_classify: bool = True
    ) -> None:
        """Add new design points to PAL instance

        Args:
            X_design (np.ndarrary): Design matrix. Two-dimensional array containing
                measurements in the rows and the features as the columns.
            classify (bool): Reclassifies the new design space, using the old model.
                This is, it runs inference, calculates the hyperrectangles, and runs
                the classification. Does not increase the iteration count.
                Note though that points that already have been classified
                as Pareto-optimal will not be re-classified,
                e.g., discarded---even if the new design points
                dominate the existing "Pareto optimal" points.
                Defaults to False.
            clean_classify (bool): Reclassifies the new design space,
                using the old model. This is, it runs inference,
                calculates the hyperrectangles, and runs the classification.
                Does not increase the iteration count.
                But, in contrast to `classify` it erases all previous classifications,
                before running the new classification. Hence, if some new design point
                dominates a previously Pareto efficient point,
                the previous Pareto optimal point will no longer be classified
                as Pareto efficient.
                This flag is incompatible with `classify`.
                If you choose `clean_classify`, PyePAL will erase
                all previous classifications,
                independent of what you choose for `classify`.
                Defaults to True.
        """

        if self.iteration <= 1:
            raise ValueError(
                "You must run a iteration before you augment the design space"
            )

        number_old_points = self.number_design_points
        number_new_points = len(X_design)

        assert isinstance(
            X_design, np.ndarray
        ), "You must provide a two-dimensional numpy array"
        assert X_design.ndim == 2, "You must provide a two-dimensional numpy array"

        if X_design.shape[1] != self.design_space.shape[1]:
            raise ValueError(
                "The design matrix you provided has shape {}, \
                    the pyepal instance uses a design matrix of shape {}.".format(
                    X_design.shape, self.design_space.shape
                )
            )

        if classify and clean_classify:
            warnings.warn(
                "You choose both `classify` and `clean_classify`. \
                PyePAL will use the `clean_classify` behavior and override \
                    all previous classifications.",
                UserWarning,
            )
            classify = False

        # Update the status matrices
        self.pareto_optimal = np.append(
            self.pareto_optimal, np.array([False] * number_new_points), 0
        )
        self.discarded = np.append(
            self.discarded, np.array([False] * number_new_points), 0
        )
        self.sampled = np.append(
            self.sampled, np.array([[False] * self.ndim] * number_new_points), 0
        )
        self.unclassified = np.append(
            self.unclassified, np.array([True] * number_new_points), 0
        )
        self.rectangle_ups = np.append(
            self.rectangle_ups, np.full([number_new_points, self.ndim], np.nan), 0
        )
        self.rectangle_lows = np.append(
            self.rectangle_lows, np.full([number_new_points, self.ndim], np.nan), 0
        )
        self.coef_var_mask = np.append(
            self.coef_var_mask, np.array([True] * number_new_points), 0
        )

        # means/std are the model predictions
        self.means = np.append(
            self.means, np.full([number_new_points, self.ndim], np.nan), 0
        )
        self.std = np.append(
            self.std, np.full([number_new_points, self.ndim], np.nan), 0
        )

        # self.y is what needs to be used for train/predict
        # as there the data has been turned into maximization
        self.y = np.append(self.y, np.zeros((number_new_points, self.ndim)), 0)

        # self._y contains the data as provided by the user
        self._y = self.y

        # measurement_uncertainity is provided in update_train_set by the user
        self.measurement_uncertainty = np.append(
            self.measurement_uncertainty, np.zeros((number_new_points, self.ndim)), 0
        )

        # Update the design space
        self.design_space = np.append(self.design_space, X_design, 0)

        new_indices = np.arange(number_old_points, self.number_design_points)

        # Make sure that the new points have the same "state" as the old ones
        # This is, we can use the new design space in a proper way for sampling
        # or classification
        # ToDo: the bug is here that iteration !=1 wherefore it will intersect
        # But this makes no sense here when we initialize the highs and lows with zeros
        if classify:
            self._predict()
            self._replace_by_measurements()
            self._update_hyperrectangles(new_indices=new_indices)
            self._classify()

        if clean_classify:
            self.pareto_optimal = np.array([False] * self.number_design_points)
            # We can be a bit more clever and only re-consider the Pareto optimal
            # points. We can leave the
            self.unclassified = np.array([True] * self.number_design_points)
            self.unclassified[self.discarded_indices] = False
            self._predict()
            self._replace_by_measurements()
            self._update_hyperrectangles(new_indices=new_indices)
            self._classify()

    def sample(
        self,
        exclude_idx: Union[np.array, None] = None,
        pooling_method: str = "fro",
        sample_discarded: bool = False,
        use_coef_var: bool = True,
    ) -> int:
        """Runs the sampling step based on the size of the hyperrectangle.
        I.e., favoring exploration.

        Args:
            exclude_idx (Union[np.array, None], optional):
                Points in design space to exclude from sampling.
                Defaults to None.
            pooling_method (str): Method that is used to aggregate
                the uncertainty in different objectives into one scalar.
                Available options are:  "fro" (Frobenius/Euclidean norm), "mean",
                "median". Defaults to "fro".
            sample_discarded (bool): if true, it will sample from all points
                and not only from the unclassified and Pareto optimal ones
            use_coef_var (bool): If True, uses the coefficient of variation instead of
                the unscaled rectangle sizes

        Raises:
            ValueError: In case there are no uncertainty rectangles,
                i.e., when the _predict has not been successfully called.

        Returns:
            int: Index of next point to evaluate in design space
        """
        if (self.rectangle_lows is None) | (self.rectangle_ups is None):
            raise ValueError(
                "You need to have uncertainty rectangles\
                     before you can peform the sampling"
            )

        sampled_mask = self.sampled_mask.copy()
        # This is to handle the case of batch sampling, where we do need
        # to make sure that we do not sample the same points
        if isinstance(exclude_idx, np.ndarray):
            if len(exclude_idx) >= 1:
                exclude_mask = np.zeros(len(sampled_mask), dtype=bool)
                exclude_mask[exclude_idx] = True

                sampled_mask += exclude_mask

        if sample_discarded:
            sampled_idx = _get_max_wt_all(
                self.rectangle_lows,
                self.rectangle_ups,
                self.means,
                sampled_mask,
                pooling_method,
                use_coef_var,
            )
        else:
            sampled_idx = _get_max_wt(
                self.rectangle_lows,
                self.rectangle_ups,
                self.means,
                self.pareto_optimal,
                self.unclassified,
                sampled_mask,
                pooling_method,
                use_coef_var,
            )

        return sampled_idx



# -*- coding: utf-8 -*-
# pylint:disable=anomalous-backslash-in-string
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Core functions for PAL"""
from typing import List, Sequence, Tuple, Union

import numpy as np
from numba import jit

from .utils import (
    dominance_check_jitted_2,
    dominance_check_jitted_3,
    is_pareto_efficient,
)

__all__: List[str] = []


def _get_uncertainty_region(  # pylint:disable=invalid-name
    mu: np.array, std: np.array, beta_sqrt: float
) -> Tuple[np.array, np.array]:
    """

    Args:
        mu (float): mean
        std (float): standard deviation
        beta_sqrt (float): scaling factor

    Returns:
        Tuple[float, float]: lower bound, upper bound
    """
    low_lim, high_lim = mu - beta_sqrt * std, mu + beta_sqrt * std
    return low_lim, high_lim


def _get_uncertainty_regions(
    mus: np.array, stds: np.array, beta_sqrt: float
) -> Union[np.array, np.array]:
    """
    Compute the lower and upper bound of the uncertainty region
         for each dimension (=target)

    Args:
        mus (np.array): means
        stds (np.array): standard deviations
        beta_sqrt (float): scaling factors

    Returns:
        Union[np.array, np.array]: lower bounds, upper bounds
    """
    low_lims, high_lims = [], []

    for i in range(0, mus.shape[1]):
        low_lim, high_lim = _get_uncertainty_region(mus[:, i], stds[:, i], beta_sqrt)
        low_lims.append(low_lim.reshape(-1, 1))
        high_lims.append(high_lim.reshape(-1, 1))

    return np.hstack(low_lims), np.hstack(high_lims)


def _union(
    lows: np.array, ups: np.array, new_lows: np.array, new_ups: np.array
) -> Union[np.array, np.array]:
    """Performing iterative intersection (eq. 6 in PAL paper) in all dimensions.

    Args:
        lows (np.array): lower bounds from previous iteration
        ups (np.array): upper bounds from previous iteration
        new_lows (np.array): lower bounds from current iteration
        new_ups (np.array): upper bounds from current iteration

    Returns:
        Union[np.array, np.array]: lower bounds, upper bounds
    """
    out_lows = []
    out_ups = []

    for i in range(0, lows.shape[1]):
        low, up = _union_one_dim(  # pylint:disable=invalid-name
            lows[:, i], ups[:, i], new_lows[:, i], new_ups[:, i]
        )
        out_lows.append(low.reshape(-1, 1))
        out_ups.append(up.reshape(-1, 1))

    out_lows_array, out_ups_array = np.hstack(out_lows), np.hstack(out_ups)

    return out_lows_array, out_ups_array


@jit(nopython=True)
def _union_one_dim(
    lows: Sequence, ups: Sequence, new_lows: Sequence, new_ups: Sequence
) -> Tuple[np.array, np.array]:
    """Used to intersect the confidence regions, for eq. 6 of the PAL paper.
    The iterative intersection ensures that all uncertainty regions
    are non-increasing with t.

    We do not check for the ordering in this function.
    We really assume that the lower limits are the lower limits
    and the upper limits are the upper limits.

    All arrays must have the same length.

    Args:
        lows (Sequence): lower bounds from previous iteration
        ups (Sequence): upper bounds from previous iteration
        new_lows (Sequence): lower bounds from current iteration
        new_ups (Sequence): upper bounds from current iteration

    Returns:
        Tuple[np.array, np.array]: array of lower limits, array of upper limits
    """
    out_lows = []
    out_ups = []

    for i, low in enumerate(lows):
        # In one dimension we can imagine the following cases where there
        # is zero intersection
        # 1) |--old range--|   |--new range--|,
        # i.e., lower new limit above old upper limit
        # 2) |--new range--|   |--old range--|,
        # i.e., upper new limit below lower old limit
        if (new_lows[i] >= ups[i]) or (new_ups[i] <= low):
            out_lows.append(new_lows[i])
            out_ups.append(new_ups[i])

        # In other cases, we want to intersect the ranges, i.e.
        # |---old range-|-|--new-range--| --> |-|
        # i.e. we take the max of the lower limits and the min of the upper limits
        else:
            out_lows.append(max(low, new_lows[i]))
            out_ups.append(min(ups[i], new_ups[i]))

    return np.array(out_lows), np.array(out_ups)


def _pareto_classify(  # pylint:disable=too-many-arguments, too-many-locals, too-many-branches
    pareto_optimal_0: np.array,
    not_pareto_optimal_0: np.array,
    unclassified_0: np.array,
    rectangle_lows: np.array,
    rectangle_ups: np.array,
    epsilon: np.array,
    is_fixed_epsilon: bool = False,
) -> Tuple[np.array, np.array, np.array]:
    """Performs the classification of the algorithm
    (p. 4 of the PAL paper, see algorithm 1/2 of the epsilon-PAL paper)

    One core concept is that once a point is classified,
    it does no longer change the class.

    When we do the comparison with +/- epsilon we always use the absolute values!
    Otherwise, we get inconcistent results depending on the sign!

    Args:
        pareto_optimal_0 (np.array): boolean mask of points classified as Pareto optimal
        not_pareto_optimal_0 (np.array): boolean mask of points
            classified as non-Pareto optimal
        unclassified_0 (np.array): boolean mask of unclassified points
        rectangle_lows (np.array): lower uncertainty boundaries
        rectangle_ups (np.array): upper uncertainty boundaries
        epsilon (np.array): granularity parameter (one per dimension)
        is_fixed_epsilon (bool): If true it assumes that epsilon contains *absolute*
            tolerance values for every objective. These would typically be calculated as
            :math:`\epsilon_i = \varepsilon_i \cdot r_i`, where :math:`r_i` is the range
            of objective :math:`i`. By default this is False. This is, we will use
            :math:`\epsilon_i = \varepsilon_i \cdot y_i` to compute the objectives,
            which hence avoids the need of knowing the range of a objective before
            using the algorithm.

    Returns:
        Tuple[list, list, list]: binary encoded list of Pareto optimal,
            non-Pareto optimal and unclassified points
    """
    pareto_optimal_t = pareto_optimal_0.copy()
    not_pareto_optimal_t = not_pareto_optimal_0.copy()
    unclassified_t = unclassified_0.copy()

    # This part is only relevant when we have points in set P
    # Then we can use those points to discard points from p_pess (P \cup U)
    if sum(pareto_optimal_0) > 0:
        pareto_indices = np.where(pareto_optimal_0)[0]
        pareto_pessimistic_lows = rectangle_lows[pareto_indices]  # p_pess(P)

        if is_fixed_epsilon:
            tolerances_0 = np.tile(epsilon, (len(pareto_pessimistic_lows), 1))
        else:
            tolerances_0 = np.abs(epsilon * pareto_pessimistic_lows)

        for i in range(0, len(unclassified_0)):
            if unclassified_t[i] == 1:
                # discard if any lower-bound epsilon dominates the upper bound
                if dominance_check_jitted_2(
                    pareto_pessimistic_lows + tolerances_0,
                    rectangle_ups[i],
                ):
                    not_pareto_optimal_t[i] = True
                    unclassified_t[i] = False

    pareto_unclassified_indices = np.where(pareto_optimal_0 | unclassified_t)[0]
    pareto_unclassified_lows = rectangle_lows[pareto_unclassified_indices]

    # assuming maximization
    pareto_unclassified_pessimistic_mask = is_pareto_efficient(
        -pareto_unclassified_lows
    )
    original_indices = pareto_unclassified_indices[pareto_unclassified_pessimistic_mask]
    pareto_unclassified_pessimistic_points = pareto_unclassified_lows[
        pareto_unclassified_pessimistic_mask
    ]

    if is_fixed_epsilon:
        tolerances_1 = np.tile(
            epsilon, (len(pareto_unclassified_pessimistic_points), 1)
        )
    else:
        tolerances_1 = epsilon * np.abs(pareto_unclassified_pessimistic_points)

    for i in range(0, len(unclassified_t)):  # pylint:disable=consider-using-enumerate
        # We can only discard points that are unclassified so far
        # We cannot discard points that are part of p_pess(P \cup U)
        if unclassified_t[i] and (i not in original_indices):
            # discard if any lower-bound epsilon dominates the upper bound
            if dominance_check_jitted_2(
                tolerances_1 + pareto_unclassified_pessimistic_points,
                rectangle_ups[i],
            ):
                not_pareto_optimal_t[i] = True
                unclassified_t[i] = False

    # now, update the pareto set
    # if there is no other point x' such that max(Rt(x')) >= min(Rt(x))
    # move x to Pareto
    unclassified_indices = np.where(unclassified_t | pareto_optimal_t)[0]
    unclassified_ups = rectangle_ups[unclassified_indices]

    index_map = {index: i for i, index in enumerate(unclassified_indices)}

    if is_fixed_epsilon:
        tolerances_2 = np.tile(epsilon, (len(rectangle_lows), 1))
    else:
        tolerances_2 = epsilon * np.abs(rectangle_lows)

    # The index map helps us to mask the current point from the unclassified_ups list
    for i in range(0, len(unclassified_t)):  # pylint:disable=consider-using-enumerate
        # again, we only care about unclassified points
        if unclassified_t[i]:
            # If there is no other point which up is epsilon dominating
            # the low of the current point,
            # the current point is epsilon-accurate Pareto optimal
            if not dominance_check_jitted_3(
                unclassified_ups,
                rectangle_lows[i] + tolerances_2[i],
                index_map[i],
            ):
                pareto_optimal_t[i] = True
                unclassified_t[i] = False

    return pareto_optimal_t, not_pareto_optimal_t, unclassified_t


@jit(nopython=True)
def _get_max_wt(  # pylint:disable=too-many-arguments
    rectangle_lows: np.array,
    rectangle_ups: np.array,
    means: np.array,
    pareto_optimal_t: np.array,
    unclassified_t: np.array,
    sampled: np.array,
    pooling_method: str = "fro",
    use_coef_var: bool = True,
) -> int:
    """Returns the index in design space with the maximum size of the hyperrectangle
    (scaled by the mean predictions, i.e., effectively,
    we use the coefficient of variation).
    Samples only from unclassified or Pareto-optimal points.

    Args:
        rectangle_lows (np.array): Lower, pessimistic, bounds of the hyperrectangles
        rectangle_ups (np.array): Upper, optimistic, bounds of the hyperrectangles
        means (np.array): Mean predictions
        pareto_optimal_t (np.array): Mask array that is True
            for the Pareto optimal points
        unclassified_t (np.array): Mask array that is True for the unclassified points
        sampled (np.array): Mask array that is True for the sampled points
        pooling_method (str): Swtich to select the pooling method with which
            the uncertainty in different objectives is aggregated into one number.
            Available options are "fro" (Frobenius/Euclidean norm), "mean", "median"
        use_coef_var (bool): If True, uses the coefficient of variation instead of
            the unscaled rectangle sizes

    Returns:
        int: index with maximum size of hyperrectangle
    """
    max_uncertainty = -np.inf
    maxid = 0

    pooling_method = pooling_method.lower()

    for i in range(0, len(unclassified_t)):  # pylint:disable=consider-using-enumerate
        # Among the points x ∈ Pt ∪ Ut, the one with the largest wt(x)
        # is chosen as the next sample xt to be evaluated.
        # Intuitively, this rule biases the sampling towards exploring,
        # and thus improving the model for, the points most likely to be Pareto-optimal.
        if ((unclassified_t[i] == 1) or (pareto_optimal_t[i] == 1)) and not sampled[
            i
        ] == 1:
            # weight is the length of the diagonal of the uncertainty region
            if use_coef_var:
                uncer = np.divide(
                    rectangle_ups[i, :] - rectangle_lows[i, :], means[i, :]
                )
            else:
                uncer = rectangle_ups[i, :] - rectangle_lows[i, :]

            uncertainty = _pool(uncer, pooling_method)
            if uncertainty > max_uncertainty:
                max_uncertainty = uncertainty
                maxid = i

    return maxid


@jit(nopython=True)
def _get_max_wt_all(  # pylint:disable=too-many-arguments
    rectangle_lows: np.array,
    rectangle_ups: np.array,
    means: np.array,
    sampled: np.array,
    pooling_method: str = "fro",
    use_coef_var: bool = True,
) -> int:
    """Returns the index in design space with the maximum size of the hyperrectangle
    (scaled by the mean predictions, i.e., effectively,
    we use the coefficient of variation).
    Samples from *all* unsampled points. This is different from the `_get_max_wt`
    function that only samples from the Pareto-optimal or unclassified points that
    have not been sampled yet.

    Args:
        rectangle_lows (np.array): Lower, pessimistic, bounds of the hyperrectangles
        rectangle_ups (np.array): Upper, optimistic, bounds of the hyperrectangles
        means (np.array): Mean predictions
        sampled (np.array): Mask array that is True for the sampled points
        pooling_method (str): Swtich to select the pooling method with which the
            uncertainty in different objectives is aggregated into one number.
            Available options are: "fro" (Frobenius/Euclidean norm), "mean", "median"
        use_coef_var (bool): If True, uses the coefficient of variation instead of
            the unscaled rectangle sizes

    Returns:
        int: index with maximum size of hyperrectangle
    """
    max_uncertainty = -np.inf
    maxid = 0

    pooling_method = pooling_method.lower()

    for i in range(0, len(sampled)):  # pylint:disable=consider-using-enumerate
        # Among the points x ∈ Pt ∪ Ut, the one with the largest wt(x)
        # is chosen as the next sample xt to be evaluated.
        # Intuitively, this rule biases the sampling towards exploring,
        # and thus improving the model for, the points most likely to be Pareto-optimal.
        if not sampled[i] == 1:
            # weight is the length of the diagonal of the uncertainty region
            if use_coef_var:
                uncer = np.divide(
                    rectangle_ups[i, :] - rectangle_lows[i, :], means[i, :]
                )
            else:
                uncer = rectangle_ups[i, :] - rectangle_lows[i, :]
            uncertainty = _pool(uncer, pooling_method)
            if uncertainty > max_uncertainty:
                max_uncertainty = uncertainty
                maxid = i

    return maxid


@jit(nopython=True)
def _pool(array: np.ndarray, method: str) -> float:
    """Summarizes an array into one number

    Args:
        array (np.ndarray): array (,1) of numbers
        method (str): Swtich to select the pooling method with which the uncertainty
            in different objectives is aggregated into one number.
            Available options are: "fro" (Frobenius/Euclidean norm), "mean", "median".
            If the string cannot be matched, it defaults to the Frobenius norm.

    Returns:
        float: result of aggregation
    """
    if method == "fro":
        return np.linalg.norm(array)
    if method == "mean":
        return np.mean(array)
    if method == "median":
        return np.median(array)
    return np.linalg.norm(array)


def _uncertainty(rectangle_ups, rectangle_lows, means):
    """Hyperrectangle sizes"""
    if (rectangle_lows is None) or (rectangle_ups is None):
        raise ValueError(
            "You must have trained a model and\
             run the prediction to calculate hyperrectangle"
        )
    coeff_var = np.nan_to_num(np.divide(rectangle_ups - rectangle_lows, means))
    uncertainty = np.linalg.norm(coeff_var, axis=1)
    return uncertainty



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model agnostic Python implementation of the epsilon-PAL algorithm"""

from setuptools import find_packages, setup

import versioneer

with open("requirements.txt", "r") as fh:
    REQUIREMENTS = fh.readlines()


with open("README.md", encoding="utf-8") as fh:
    LONG_DESCRIPTION = fh.read()

gpy_requirements = [
    "GPy~=1.9.9",
    "matplotlib==3.*",
]
gbdt_requirements = ["lightgbm==3.*"]
neural_tangents_requirements = ["neural_tangents==0.*", "jaxlib==0.*"]
setup(
    name="pyepal",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="PyePAL implemented the epsilon-PAL active learning algorithm",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(include=["pyepal", "pyepal.*"]),
    url="https://github.com/kjappelbaum/PyePAL",
    license="Apache 2.0",
    install_requires=REQUIREMENTS,
    extras_require={
        "testing": ["pytest==6.*", "pytest-cov==2.*"],
        "docs": [
            "sphinx==3.*",
            "sphinx-book-theme==0.*",
            "sphinx-autodoc-typehints==1.*",
            "sphinx-copybutton==0.*",
        ],
        "pre-commit": [
            "pre-commit==2.*",
            "pylint==2.*",
            "isort==5.*",
        ],
        "dev": [
            "versioneer==0.*",
            "black==20.*",
        ],
        "GPy": gpy_requirements,
        "GBDT": gbdt_requirements,
        "neural_tangents": neural_tangents_requirements,
        "all": neural_tangents_requirements + gbdt_requirements + gpy_requirements,
    },
    author="PyePAL authors",
    author_email="kevin.jablonka@epfl.ch, brian.yoo@basf.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for dealing with Pareto fronts in general"""

import numpy as np
from numba import jit
from scipy.spatial import distance
from sklearn import metrics
from sklearn.cluster import KMeans

from ._hypervolume import HypervolumeIndicator

__all__ = [
    "dominance_check",
    "dominance_check_jitted",
    "dominance_check_jitted_2",
    "dominance_check_jitted_3",
    "is_pareto_efficient",
    "exhaust_loop",
    "get_kmeans_samples",
    "get_maxmin_samples",
    "get_hypervolume",
]


@jit(nopython=True)
def dominance_check(point1, point2) -> bool:
    """One point dominates another if it is not worse in all objectives
    and strictly better in at least one. This here assumes we want to maximize"""
    if np.all(point1 >= point2) and np.any(point1 > point2):
        return True

    return False


@jit(nopython=True)
def dominance_check_jitted(point: np.array, array: np.array) -> bool:
    """Check if point dominates any point in array"""
    arr_sorted = array[array[:, 0].argsort()]
    for i in range(len(arr_sorted)):  # pylint:disable=consider-using-enumerate
        if dominance_check(point, arr_sorted[i]):
            return True
    return False


@jit(nopython=True)
def dominance_check_jitted_2(array: np.array, point: np.array) -> bool:
    """Check if any point in array dominates point"""
    arr_sorted = array[array[:, 0].argsort()[::-1]]
    for i in range(len(arr_sorted)):  # pylint:disable=consider-using-enumerate
        if dominance_check(arr_sorted[i], point):
            return True
    return False


@jit(nopython=True)
def dominance_check_jitted_3(array: np.array, point: np.array, ignore_me: int) -> bool:
    """Check if any point in array dominates point. ignore_me
    since numba does not understand masked arrays"""
    sorted_idx = array[:, 0].argsort()[::-1]
    ignore_me = np.where(sorted_idx == ignore_me)[0][0]
    arr_sorted = array[sorted_idx]
    for i in range(len(arr_sorted)):  # pylint:disable=consider-using-enumerate
        if i != ignore_me:
            if dominance_check(arr_sorted[i], point):
                return True
    return False


def is_pareto_efficient(costs: np.array, return_mask: bool = True) -> np.array:
    """Find the Pareto efficient points
    Based on https://stackoverflow.com/questions/
    32791911/fast-calculation-of-pareto-front-in-python

    Args:
        costs (np.array): An (n_points, n_costs) array
        return_mask (bool, optional): True to return a mask,
            Otherwise it will be a (n_efficient_points, ) integer array of indices.
            Defaults to True.

    Returns:
        np.array: [description]
    """
    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):
        nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype=bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    return is_efficient


def exhaust_loop(
    palinstance, y: np.array, batch_size: int = 1
):  # pylint:disable=invalid-name
    """Helper function that takes an initialized PAL instance
    and loops the sampling until there is no unclassified point left.
    This is useful if all measurements are already taken and one
    wants to test the algorithm with different hyperparameters.

    Args:
        palinstance (PALBase): A initialized instance of a class that
            inherited from PALBase and implemented the ._train() and
            ._predict() functions
        y (np.array): Measurements.
            The number of measurements must equal the number of
            points in the design space.
        batch_size (int, optional): Number of indices that will be returned.
                Defaults to 10.

    Returns:
        None. The PAL instance is updated in place
    """
    assert palinstance.number_design_points == len(
        y
    ), "The number of points in the design space must equal the number of measurements"
    while sum(palinstance.unclassified):
        idx = palinstance.run_one_step(batch_size=batch_size)
        if idx is not None:
            palinstance.update_train_set(idx, y[idx])


def get_kmeans_samples(  # pylint:disable=invalid-name
    X: np.array, n_samples: int, **kwargs
) -> np.array:
    """Get the samples that are closest to the k=n_samples centroids

    Args:
        X (np.array): Feature array, on which the KMeans clustering is run
        n_samples (int): number of samples are should be selected
        **kwargs passed to the KMeans

    Returns:
        np.array:  selected_indices
    """

    assert (
        len(X) > n_samples
    ), "The numbers of points that shall be selected (n_samples),\
         needs to be smaller than the length of the feature matrix (X)"

    assert n_samples > 0 and isinstance(
        n_samples, int
    ), "The number of points that shall be selected (n_samples)\
             needs to be an integer greater than 0"

    kmeans = KMeans(n_samples, **kwargs).fit(X)
    cluster_centers = kmeans.cluster_centers_
    closest, _ = metrics.pairwise_distances_argmin_min(cluster_centers, X)

    return closest


def get_maxmin_samples(  # pylint:disable=invalid-name
    X: np.array,
    n_samples: int,
    metric: str = "euclidean",
    init: str = "mean",
    seed: int = None,
    **kwargs
) -> np.array:
    """Greedy maxmin sampling, also known as Kennard-Stone sampling (1).
    Note that a greedy sampling is not guaranteed to give the ideal solution
    and the output will depend on the random initialization (if this is chosen).

    If you need a good solution, you can restart this algorithm multiple times
    with random initialization and different random seeds
    and use a coverage metric to quantify how well
    the space is covered. Some metrics are described in (2). In contrast to the
    code provided with (2) and (3) we do not consider the feature importance for the
    selection as this is typically not known beforehand.

    You might want to standardize your data before applying this sampling function.

    Some more sampling options are provided in our structure_comp (4) Python package.
    Also, this implementation here is quite memory hungry.

    References:
    (1) Kennard, R. W.; Stone, L. A. Computer Aided Design of Experiments.
    Technometrics 1969, 11 (1), 137–148. https://doi.org/10.1080/00401706.1969.10490666.
    (2) Moosavi, S. M.; Nandy, A.; Jablonka, K. M.; Ongari, D.; Janet,
    J. P.; Boyd, P. G.; Lee, Y.; Smit, B.; Kulik, H. J.
    Understanding the Diversity of the Metal-Organic Framework Ecosystem.
    Nature Communications 2020, 11 (1), 4068.
    https://doi.org/10.1038/s41467-020-17755-8.
    (3) Moosavi, S. M.; Chidambaram, A.; Talirz, L.; Haranczyk, M.; Stylianou, K. C.;
    Smit, B. Capturing Chemical Intuition in Synthesis of Metal-Organic Frameworks.
    Nat Commun 2019, 10 (1), 539. https://doi.org/10.1038/s41467-019-08483-9.
    (4) https://github.com/kjappelbaum/structure_comp

    Args:
        X (np.array): Feature array, this is the array
            that is used to perform the sampling

        n_samples (int): number of points that will be selected,
            needs to be lower than the length of X

        metric (str, optional): Distance metric to use for the maxmin calculation.
            Must be a valid option of scipy.spatial.distance.cdist
            (‘braycurtis’, ‘canberra’, ‘chebyshev’,
            ‘cityblock’, ‘correlation’, ‘cosine’, ‘dice’,
            ‘euclidean’, ‘hamming’, ‘jaccard’, ‘jensenshannon’, ‘kulsinski’,
            ‘mahalanobis’, ‘matching’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’,
            ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’,
            ‘wminkowski’, ‘yule’). Defaults to 'euclidean'

        init (str, optional): either 'mean', 'median', or 'random'.
            Determines how the initial point is chosen. Defaults to 'center'

        seed (int, optional): seed for the random number generator. Defaults to None.
        **kwargs passed to the cdist

    Returns:
        np.array: selected_indices
    """
    np.random.seed(seed)

    assert (
        len(X) > n_samples
    ), "The numbers of points that shall be selected (n_samples),\
         needs to be smaller than the length of the feature matrix (X)"

    assert n_samples > 0 and isinstance(
        n_samples, int
    ), "The number of points that shall be selected (n_samples)\
         needs to be an integer greater than 0"

    greedy_data = []

    if init == "random":
        index = np.random.randint(0, len(X) - 1)
    elif init == "mean":
        index = np.argmin(np.linalg.norm(X - np.mean(X, axis=0), axis=1))
    else:
        index = np.argmin(np.linalg.norm(X - np.median(X, axis=0), axis=1))
    greedy_data.append(X[index])
    remaining = np.delete(X, index, 0)
    while len(greedy_data) < n_samples:
        dist = distance.cdist(remaining, greedy_data, metric, **kwargs)
        greedy_index = np.argmax(np.min(dist, axis=0))
        greedy_data.append(remaining[greedy_index])
        remaining = np.delete(remaining, greedy_index, 0)

    greedy_indices = []

    for datum in greedy_data:
        greedy_indices.append(np.array(np.where(np.all(X == datum, axis=1)))[0])

    greedy_indices = np.concatenate(greedy_indices).ravel()

    return greedy_indices


def get_hypervolume(
    pareto_front: np.array, reference_vector: np.array, prefactor: float = -1
) -> float:
    """Compute the hypervolume indicator of a Pareto front
    I multiply it with minus one as we assume that we want
    to maximize all objective and then we calculate the area

    f1
    |
    |----|
    |     -|
    |       -|
    ------------ f2

    But the code we use for the hv indicator assumes that the reference vector
    is larger than all the points in the Pareto front.
    For this reason, we then flip all the signs using prefactor

    This indicator is not needed for the epsilon-PAL algorithm itself
    but only to allow tracking a metric that might help the user to see
    if the algorithm converges.
    """
    hv_instance = HypervolumeIndicator(reference_vector)
    volume = hv_instance.compute(pareto_front * prefactor)
    return volume


@jit(nopython=True)
def get_nondimensional_pareto_error(
    y_true: np.ndarray, y_pred: np.ndarray, ranges: np.ndarray
) -> float:
    """Calculates a non-dimensional error metric,
    the scaled minimum maximum average distance of a Pareto-optimal
    point to one in the predicted set. This metric is typically used
    if the full objective space is known, i.e., when we know all y_true
    (the true Pareto front) and when we can calculate the ranges based on
    the full objective space. A use case can be to compare runs with different
    epsilon or with different algorithms in a "fairer" way as no choice of
    a hypervolume reference point is needed.

    Args:
        y_true (np.ndarray): True Pareto front
        y_pred (np.ndarray): Predicted Pareto front
        ranges (np.ndarray): Range of every objective, typically calculated
            based on the full objective space

    Returns:
        float: error metric
    """
    cum_distance = 0
    for i in range(len(y_true)):  # pylint:disable=consider-using-enumerate
        maximum_distance = np.inf
        for j in range(len(y_pred)):  # pylint:disable=consider-using-enumerate
            current_maximum_distance = np.max(np.abs(y_true[i] - y_pred[j]) / ranges)
            if current_maximum_distance < maximum_distance:
                maximum_distance = current_maximum_distance

        cum_distance += maximum_distance

    return cum_distance / len(y_true)



# -*- coding: utf-8 -*-
# Copyright 2020 PyePAL authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PAL using GPy GPR models"""
import concurrent.futures
from functools import partial

import numpy as np

from .pal_base import PALBase
from .schedules import linear
from .validate_inputs import validate_njobs, validate_number_models

__all__ = ["PALGPy"]


def _train_model_picklable(i, models, restarts):
    model = models[i]
    model.optimize_restarts(restarts)
    return model


class PALGPy(PALBase):
    """PAL class for a list of GPy GPR models, with one model per objective"""

    def __init__(self, *args, **kwargs):
        """Contruct the PALGPy instance

        Args:
            X_design (np.array): Design space (feature matrix)
            models (list): Machine learning models
            ndim (int): Number of objectives
            epsilon (Union[list, float], optional): Epsilon hyperparameter.
                Defaults to 0.01.
            delta (float, optional): Delta hyperparameter. Defaults to 0.05.
            beta_scale (float, optional): Scaling parameter for beta.
                If not equal to 1, the theoretical guarantees do not necessarily hold.
                Also note that the parametrization depends on the kernel type.
                Defaults to 1/9.
            goals (List[str], optional): If a list, provide "min" for every objective
                that shall be minimized and "max" for every objective
                that shall be maximized. Defaults to None, which means
                that the code maximizes all objectives.
            coef_var_threshold (float, optional): Use only points with
                a coefficient of variation below this threshold
                in the classification step. Defaults to 3.
            restarts (int): Number of random restarts that are used for hyperparameter
                optimization. Defaults to 20.
            n_jobs (int): Number of parallel processes that are used to fit
                the GPR models. Defaults to 1.
        """
        from .validate_inputs import (  # pylint:disable=import-outside-toplevel
            validate_gpy_model,
        )

        self.restarts = kwargs.pop("restarts", 20)
        self.n_jobs = validate_njobs(kwargs.pop("n_jobs", 1))

        assert isinstance(
            self.restarts, int
        ), "the restarts keyword must be of type int"
        super().__init__(*args, **kwargs)

        validate_number_models(self.models, self.ndim)
        validate_gpy_model(self.models)

    def _set_data(self):
        for i, model in enumerate(self.models):
            model.set_XY(
                self.design_space[self.sampled[:, i]],
                self.y[self.sampled[:, i], i].reshape(-1, 1),
            )

    def _train(self):
        pass  # There is no training in instance based models

    def _predict(self):

        from ..models.gpr import predict  # pylint:disable=import-outside-toplevel

        means, stds = [], []
        for model in self.models:
            mean, std = predict(model, self.design_space)
            means.append(mean.reshape(-1, 1))
            stds.append(std.reshape(-1, 1))

        self.means = np.hstack(means)
        self.std = np.hstack(stds)

    def _set_hyperparameters(self):
        models = []

        train_model_pickleable_partial = partial(
            _train_model_picklable, models=self.models, restarts=self.restarts
        )
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.n_jobs
        ) as executor:
            for model in executor.map(train_model_pickleable_partial, range(self.ndim)):
                models.append(model)
        self.models = models

    def _should_optimize_hyperparameters(self) -> bool:
        return linear(self.iteration, 10)
